import os
import pandas as pd
from tqdm import tqdm
from glob import glob
import matplotlib.pyplot as plt
import svgutils.compose as sc
import numpy as np
import pickle as pkl
import seaborn as sns
from dlab.sglx_analysis import readAPMeta
from dlab import sorting_quality as sq
from matplotlib.gridspec import GridSpec

"""
Use of all functions in this file are based on a somewhat standardized data format unique to Juan Santiago's analysis pipeline. Data for 
matrices are stored in a dictionary containg array data (frames) and frame times. Data for gratings stimuli are stored in a large 
compiled dataframe. Imported dlab functions are sourced from the denmanlab Github repository
"""

#generates spike triggered average data used to approximate linear receptive fields into a numpy array format
#TODO add more flexibility for tau values. Currently only able to adjust number but not range
def rf_array(spike_data, probe, unit, stim_data, stim_times,num_taus=10):
    probe_spikes = spike_data[spike_data['probe']==probe]
    spiketimes = np.array(probe_spikes['times'][unit])
    taus = np.round(np.linspace(-0.01,0.28,num_taus),2)
    srf_list=[]
    
    for k,tau in tqdm(enumerate(taus)):
        idx_arr = []
        stimspikes = []
        spiketime_adj = spiketimes-tau
        for i in range(spiketimes.shape[0]):
            bool_idx = np.logical_and(spiketime_adj[i] > stim_times[0] , spiketime_adj[i] < stim_times[-1])
            idx_arr.append(bool_idx)
        spike_adj_stim = spiketime_adj[idx_arr]

        frame_index = []
        for j in range(spike_adj_stim.shape[0]):
            stim_frame = np.min(np.where(stim_times>spike_adj_stim[j]))-1
            frame_index.append(stim_frame)    

        srf_frames = stim_data[frame_index]
        srf = srf_frames.mean(axis=(0))
        srf_list.append(srf)

    output = dict(zip(taus,srf_list))
    return(output)

#Takes the array output from rf_array and puts out a nicely formatted grid of imshow heatmaps
def rf_grid(data, colormap ='viridis',interp='none',title='',nrows=5,ncols=2):
    fig,axs = plt.subplots(ncols, nrows,figsize=(10,10))
    axs = axs.ravel()
    taus = np.round(np.linspace(-0.01,0.28,len(data)),2)
    for i,j in tqdm(enumerate(taus)):
        data2 = data[j]
        out_mean = np.mean(data2)
        out_std = np.std(data2)
        col_min = out_mean-(out_std*3)
        col_max = out_mean+(out_std*3)
        #Generate subplots
        axs[i].imshow(data2, clim=(col_min,col_max),cmap=colormap, 
                      interpolation=interp
                     )
        axs[i].set_title(j)
        for ax in fig.get_axes():
            ax.set_frame_on(False);
            ax.set_xticklabels('',visible=False);
            ax.set_xticks([]);
            ax.set_yticklabels('',visible=False);
            ax.set_yticks([])
            ax.set_aspect(1.0)
            ax.set_xlim(0,64)
            ax.set_ylim(0,64)
        plt.tight_layout()
        plt.suptitle(title,y=0.75)
    return(fig)

#takes spike data and triggers and yields overlaid psth ribbon plots with variance
#TODO add colormap options
#TODO separate out calculations into their own function to simplify both plotting functions
def psth_line_overlay(spike_data, probe, unit, stim_data, condition, title='', pre=0.5, post=2.5,binsize=0.05,variance=True):
    spike_data = spike_data[spike_data['probe']==probe]
    times = np.array(spike_data.times[unit])
    numbins = int((post+pre)/binsize)
    conds = np.unique(stim_data[condition])
    num_conds = len(conds)
    x = np.arange(-pre,post,binsize)
    colors = plt.cm.viridis(np.linspace(0,1,num_conds))
    
    psth_all=[]
    fig,ax = plt.subplots()
    
    for i,cond in enumerate(np.unique(stim_data[condition])):
        triggers = np.array(stim_data['times'][stim_data[condition] == cond])
        bytrial = np.zeros((len(triggers),numbins-1))
        for j, trigger in enumerate(triggers):
            trial = triggers[j]
            start = trial-pre
            end = trial+post
            bins_ = np.arange(start,end,binsize)
            trial_spikes = times[np.logical_and(times>=start, times<=end)]
            hist,edges = np.histogram(trial_spikes,bins=bins_)
            if len(hist)==numbins-1:
                bytrial[j]=hist
            elif len(hist)==numbins:
                bytrial[j]=hist[:-1]
        psth = np.mean(bytrial,axis=0)/binsize
        if isinstance(conds[i],float)==True:
            ax.plot(x[:-1],psth, color=colors[i], label=str(round(conds[i],2)))
#         if isinstance(conds[i],tuple)==True:
#             ax.plot(x[:-1],psth, color=colors[i], label=str(round(conds[i],2)))
        else:
            ax.plot(x[:-1],psth, color=colors[i], label=str(conds[i]))
        if variance == True:
            var = np.std(bytrial,axis=0)/binsize/np.sqrt((len(triggers)))
            upper = psth+var
            lower = psth-var
            ax.fill_between(x[:-1],upper,psth,alpha=0.1,color=colors[i])
            ax.fill_between(x[:-1],lower,psth,alpha=0.1,color=colors[i])
    ax.axvline(0,linestyle='dashed')
    plt.legend(loc=(1.05,0.48))
    plt.title(title)
    plt.gca().spines['top'].set_visible(False)
    plt.gca().spines['right'].set_visible(False)
    plt.draw()
    return(fig)

#similar to psth_line, this function yields psth plots for gratings stimuli as thin bars with color intensiity indicating neuron response 
#in Hz. These plots, while superior in organization and readability provide only a weak indication of baseline firing rate and lack display
#of variance
#TODO Add colormap options
def psth_bars(spike_data,probe,unit,stim_data,condition,title='',pre=0.5,post=2.5,binsize=0.05):
    spike_data = spike_data[spike_data['probe']==probe]
    times = np.array(spike_data.times[unit])
    numbins = int((post+pre)/binsize)
    num_conds = len(np.unique(stim_data[condition]))
    
    psth_all=[]
    fig,ax = plt.subplots(num_conds,1)
    
    for i,cond in enumerate(np.unique(stim_data[condition])):
        triggers = np.array(stim_data['times'][stim_data[condition] == cond])
        bytrial = np.zeros((len(triggers),numbins-1))
        for j, trigger in enumerate(triggers):
            trial = triggers[j]
            start = trial-pre
            end = trial+post
            bins_ = np.arange(start,end,binsize)
            trial_spikes = times[np.logical_and(times>=start, times<=end)]
            hist,edges = np.histogram(trial_spikes,bins=bins_)
            if len(hist)==numbins-1:
                bytrial[j]=hist
            elif len(hist)==numbins:
                bytrial[j]=hist[:-1]
        psth = np.mean(bytrial,axis=0)/binsize
        psth = np.reshape(psth,(1,len(psth)))
        psth_all.append(psth)
        im = ax[i].imshow(psth,aspect=3, vmin=0,vmax=np.max(psth_all),interpolation='gaussian')
        if isinstance(cond,float)==True:
            ax[i].set_ylabel(str(round(cond,2)),rotation=0,labelpad=20,fontsize=12,va='center')
        #elif isinstance(cond,tuple)==True:
            ax[i].set_ylabel(str(cond),rotation=0,labelpad=32,fontsize=10,va='center')
        else:
            ax[i].set_ylabel(str(cond),rotation=0,labelpad=20,fontsize=12,va='center')
        ax[i].set_yticks([])
        ax[i].axvline(pre/binsize,color='r',linewidth=3)
        ax[i].set_xticks([])

    fig.subplots_adjust(hspace=0,wspace=0)
    cbar_ax = fig.add_axes([0.93, 0.15, 0.03, 0.7])
    fig.suptitle(title, fontsize=15)
    fig.colorbar(im,cax=cbar_ax)
    cbar_ax.set_title('Hz')
    plt.show()
    return(fig)


#For a given neuron in a given dataset, this function yields plots for all stimuli responses of interest. Because of how matrix data/times
#are structured, the inputs for those arguments must be structured as a list. 
#e.g. mtx_data = [matrix_data['luminance_data'],matrix_data['green_data'],matrix_data['uv_data']]
#     mtx_times = [matrix_data['luminance_times'],matrix_data['color_times'],matrix_data['color_times']]
#
#TODO Adjust color options across functions to allow for unified color palette
#TODO Add flexibility to the folder/directory creation portion so the function doesn't fail when the folder exists
#TODO Add flexibility to stimulus inputs since sometimes stimulus fails during experiment
#TODO Add ability to use alternative save locations
def single_cell_summary(path,probe,spike_data,unit,matrix_data,matrix_times,gratings_data,
                        gratings_conditions = ['orientation','color','green','uv'],psth = 'lines',colors='viridis',save=False
                        ):
    
    raw_data = glob(path+'*'+str(probe)+'\*ap.bin')[0]
    imec_meta = readMeta(glob(path+'*'+str(probe))[0])
    sampRate = float(imec_meta['imSampRate']) #get sampling rate (Hz)
    if os.path.isdir(os.path.join(path,'Figures'))==False:
        os.mkdir(os.path.join(path,'Figures'))
        figure_path = os.path.join(path,'Figures')
    else:
        figure_path = os.path.join(path,'Figures')
    if os.path.isdir(os.path.join(path,'Figures',probe))==False:
        os.mkdir(os.path.join(path,'Figures',probe))
        probe_folder = os.path.join(path,'Figures',probe)
    else:
        probe_folder = os.path.join(path,'Figures',probe)
    if os.path.isdir(os.path.join(path,'Figures',probe,'unit'+str(unit)))==False:
        os.mkdir(os.path.join(path,'Figures',probe,'unit'+str(unit)))
        image_path = os.path.join(path,'Figures',probe,'unit'+str(unit))
    else:
       image_path = os.path.join(path,'Figures',probe,'unit'+str(unit)) 
    
    #Separate units by probe
    units_df = spike_data.loc[spike_data['probe']==probe]
    
    #generate waveform figure
    mean_wf = sq.mean_waveform(rawdata=raw_data,times=units_df.times[unit],channels=385,sampling_rate=float(sampRate))
    for i in mean_wf:
        plt.plot(i)
        plt.gca().spines['top'].set_visible(False)
        plt.gca().spines['right'].set_visible(False)
        plt.xticks([])
        plt.savefig(os.path.join(image_path,'unit'+str(unit)+'_waveform.svg'))
        plt.savefig(os.path.join(image_path,'unit'+str(unit)+'_waveform.jpg'))
        
    #generate receptive field grid figures
    lum_rfs = rf_array(units_df, probe, unit, matrix_data[0], matrix_times[0])
    green_rfs = rf_array(units_df, probe, unit, matrix_data[1], matrix_times[1])
    uv_rfs = rf_array(units_df, probe, unit, matrix_data[2], matrix_times[2])
    
    rf_outputs = [lum_rfs,green_rfs, uv_rfs]
    rf_names = ['Luminance','Green','UV']
    
    for num,j in enumerate(rf_outputs):
        grid_fig = rf_grid(j,title=rf_names[num])
        grid_fig.savefig(os.path.join(image_path,'unit'+str(unit)+'_'+str(num)+'_grid.svg'))
        grid_fig.savefig(os.path.join(image_path,'unit'+str(unit)+'_'+str(num)+'_grid.jpg'))
        
    color_condition = []
    for x in range(gratings_data.shape[0]):
        b = round(gratings_data['uv'][x],2)
        g = round(gratings_data['green'][x],2)
        condition = tuple([g, b])
        color_condition.append(condition)
        
    gratings_data['color']=color_condition
    
    for k in gratings_conditions:
        gratings_df = gratings_data[gratings_data['condition']==k]
        gratings_df = gratings_df.drop(columns=['index']).reset_index()

        if k == 'color':
            if psth == 'lines':
                #Generate PSTH for evoked response to green, uv, and combined color condition   
                psth_color = psth_line_overlay(spike_data=units_df,probe=str(probe),unit=unit,stim_data=gratings_df,condition='color',
                                               title = 'Color Condition')
                psth_green = psth_line_overlay(spike_data=units_df,probe=str(probe),unit=unit,stim_data=gratings_df,condition='green',
                                  title = 'Green')
                psth_uv = psth_line_overlay(spike_data=units_df,probe=str(probe),unit=unit,stim_data=gratings_df,condition='uv',
                                  title = 'UV')
                if save==True:
                    psth_color.savefig(os.path.join(image_path,'unit'+str(unit)+'_psthColor.svg'))
                    psth_color.savefig(os.path.join(image_path,'unit'+str(unit)+'_psthColor.jpg'))
                    psth_green.savefig(os.path.join(image_path,'unit'+str(unit)+'_psthGreen.svg'))
                    psth_green.savefig(os.path.join(image_path,'unit'+str(unit)+'_psthGreen.jpg'))
                    psth_uv.savefig(os.path.join(image_path,'unit'+str(unit)+'_psthUV.svg'))
                    psth_uv.savefig(os.path.join(image_path,'unit'+str(unit)+'_psthUV.jpg'))


            if psth == 'heatmap':
                psth_color = psth_bars(spike_data=units_df,probe=str(probe),unit=unit,stim_data = gratings_df,condition='color',
                          title = 'Color Condition')
                psth_green = psth_bars(spike_data=units_df,probe=str(probe), unit=unit,stim_data = gratings_df,condition='green',
                          title = 'Green')
                psth_uv = psth_bars(spike_data=units_df,probe=str(probe), unit=unit,stim_data = gratings_df,condition='uv',
                          title = 'UV')
                if save==True:
                    psth_color.savefig(os.path.join(image_path,'unit'+str(unit)+'_psthColor.svg'))
                    psth_color.savefig(os.path.join(image_path,'unit'+str(unit)+'_psthColor.jpg'))
                    psth_green.savefig(os.path.join(image_path,'unit'+str(unit)+'_psthGreen.svg'))
                    psth_green.savefig(os.path.join(image_path,'unit'+str(unit)+'_psthGreen.jpg'))
                    psth_uv.savefig(os.path.join(image_path,'unit'+str(unit)+'_psthUV.svg'))
                    psth_uv.savefig(os.path.join(image_path,'unit'+str(unit)+'_psthUV.jpg'))                   
                
        else:
            if psth == 'lines':
                psth_ori = psth_line_overlay(spike_data = units_df,probe=str(probe), unit = unit, stim_data = gratings_df, condition='ori',
                                  title = 'Orientation')
                psth_sf = psth_line_overlay(spike_data = units_df,probe=str(probe), unit = unit, stim_data = gratings_df, condition='SF',
                                  title = 'Spatial Frequency')
                if save==True:
                    psth_ori.savefig(os.path.join(image_path,'unit'+str(unit)+'_psthOri_'+k+'.svg'))
                    psth_ori.savefig(os.path.join(image_path,'unit'+str(unit)+'_psthOri_'+k+'.jpg'))
                    psth_sf.savefig(os.path.join(image_path,'unit'+str(unit)+'_psthSF_'+k+'.jpg'))

            if psth == 'heatmap':
                psth_ori = psth_bars(spike_data = units_df,probe=str(probe), unit = unit, stim_data = gratings_df, condition='ori',
                                  title = 'Orientation')
                psth_sf = psth_bars(spike_data = units_df,probe=str(probe), unit = unit, stim_data = gratings_df, condition='SF',
                                  title = 'Spatial Frequency')
                if save==True:
                    psth_ori.savefig(os.path.join(image_path,'unit'+str(unit)+'_psthOri_'+k+'.svg'))
                    psth_ori.savefig(os.path.join(image_path,'unit'+str(unit)+'_psthOri_'+k+'.jpg'))
                    psth_sf.savefig(os.path.join(image_path,'unit'+str(unit)+'_psthSF_'+k+'.svg'))
                    psth_sf.savefig(os.path.join(image_path,'unit'+str(unit)+'_psthSF_'+k+'.jpg'))
    if save==True:
        print('Plots have been saved to '+image_path)


#Far from done. Still needs a lot of work to properly format polished figure. 
#TODO Arrange figure into panels with titles to denote each stimulus
#TODO Add save option
#TODO Engineer flexibility for different/missing stimuli
def summary_plot(img_path,mouse_id,experiment,save=False):
    images = glob(img_path+'\*.svg')
    receptive_fields=[]
    gratings_color=[]
    gratings_orientation=[]
    gratings_green=[]
    gratings_uv=[]
    
    for i in images:
        if 'waveform' in i:
            waveform = i
        if 'grid' in i:
            receptive_fields.append(i)
        if 'psthColor' in i:
            gratings_color.append(i)
        if 'psthGreen' in i:
            gratings_color.append(i)
        if 'psthUV' in i:
            gratings_color.append(i)
        if '_ori' in i: 
            gratings_orientation.append(i)
        if '_green' in i:
            gratings_green.append(i)
        if '_uv' in i: 
            gratings_uv.append(i)\
            
    a = sc.Figure('25cm','200cm',
                    sc.SVG(waveform),
                    sc.SVG(receptive_fields[0]),
                    sc.SVG(receptive_fields[1]),
                    sc.SVG(receptive_fields[2]),
                    sc.SVG(gratings_color[0]),
                    sc.SVG(gratings_color[1]),
                    sc.SVG(gratings_color[2]),
                    sc.SVG(gratings_orientation[0]),
                    sc.SVG(gratings_orientation[1]),
                    sc.SVG(gratings_green[0]),
                    sc.SVG(gratings_green[1]),
                    sc.SVG(gratings_uv[0]),
                    sc.SVG(gratings_uv[1])).tile(1,13)       
    return(a)          


##imports===============================================================================
from scipy.spatial.distance import cdist
import numpy as np
import os, sys,glob, copy, csv
import matplotlib.pyplot as plt
import matplotlib.patches as patches
from matplotlib.colors import Normalize
from  dlab.continuous_traces import get_channel_count, filtr
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis as LDA
from sklearn.manifold import TSNE
from scipy.cluster.vq import kmeans2
##===============================================================================





##===============================================================================
#probe geometry, for the summary figure only
option234_xpositions = np.zeros((192,2))
option234_ypositions = np.zeros((192,2))
option234_positions = np.zeros((383,2))
option234_positions[:,0][::4] = 21
option234_positions[:,0][1::4] = 53
option234_positions[:,0][2::4] = 5
option234_positions[:,0][3::4] = 37
option234_positions[:,1] = np.floor(np.linspace(383,0,383)/2) * 20
##===============================================================================

##helper functions===============================================================================
def read_kilosort_params(filename):
    f=open(filename)
    params = {}
    for line in list(f):
        try:
            params[line.split(' =')[0]]=line.split('= ')[1].replace('\r\n','')
        except:
            pass
    return params

def read_cluster_groups_CSV(directory):
    cluster_id = [];
    if os.path.isfile(os.path.join(directory,'cluster_group.tsv')):
        cluster_id = [row for row in csv.reader(open(os.path.join(directory,'cluster_group.tsv')))];
    else:
        if os.path.isfile(os.path.join(directory,'cluster_groups.csv')):
            cluster_id = [row for row in csv.reader(open(os.path.join(directory,'cluster_groups.csv')))];
        else:
            print('could not find cluster groups csv or tsv')
            return None
    good=[];mua=[];unsorted=[]
    for i in np.arange(1,np.shape(cluster_id)[0]):
        if cluster_id[i][0].split('\t')[1] == 'good':#if it is a 'good' cluster by manual sort
            good.append(cluster_id[i][0].split('\t')[0])
        if cluster_id[i][0].split('\t')[1] == 'mua':#if it is a 'good' cluster by manual sort
            mua.append(cluster_id[i][0].split('\t')[0])
        if cluster_id[i][0].split('\t')[1] == 'unsorted':#if it is a 'good' cluster by manual sort
            unsorted.append(cluster_id[i][0].split('\t')[0])
    return (np.array(good).astype(int),np.array(mua).astype(int),np.array(unsorted).astype(int))
    
def count_unique(x):
    values=[]
    instances=[]
    for v in np.unique(x):
        values.extend([v])
        instances.extend([len(np.where(np.array(x)==v)[0].flatten())])
    return values, instances

def ismember(a, b):
    bind = {}
    for i, elt in enumerate(b):
        if elt not in bind:
            bind[elt] = i
    return [bind.get(itm, None) for itm in a]  # None can be replaced by any other "not in b" 

def load_waveforms(data,channel,times,pre=0.5,post=1.5,channels=384,sampling_rate=30000):
    #input can be memory mapped file or a string specifiying the file to memory map. if string, deletes the memory mapped object when done, for hygiene.
    pre = pre * .001
    post = post * .001
    channel = int(channel)
    channels = int(channels)
    if type(data)==str:
        mm = np.memmap(data, dtype=np.int16, mode='r')
    else:
        mm=data
    waveforms=[]
    for i in times:
        start = int((i - pre) * sampling_rate) * int(channels)
        temp = mm[start:start+int((pre+post)*sampling_rate*channels)][channel::channels]# - mm[start:start+int((pre+post)*sampling_rate*channels)][channel::channels][0]
        temp = temp - temp[0]
        waveforms.extend([temp * 0.195])
    if type(data)==str:
        del mm
    return waveforms

def mean_waveform(rawdata,times,pre=0.5,post=1.5,channels=384,sampling_rate=30000):
    mean_waveform = []#np.zeros(channels,int((pre+post)*.001)*sampling_rate*channels)
    for i,ch in enumerate(np.linspace(0,channels-1,channels).astype(int)):
        try:
            w = load_waveforms(rawdata,ch,times,pre,post,channels,sampling_rate)
        except:
            w = np.zeros((len(times),int(((pre+post)/1000.)*sampling_rate)))
        mean_waveform.append(np.mean(w,axis=0))#[i,:]=np.mean(w).flatten()
    return mean_waveform

def probe_waveforms(rawdata,times,pre=0.5,post=1.5,channels=384,sampling_rate=30000):
    probe_waveform = np.zeros((channels,int((pre+post)*.001*sampling_rate),len(times)))
    for i,ch in enumerate(np.linspace(0,channels-1,channels).astype(int)):
        try:
            probe_waveform[i,:,:] = np.array(load_waveforms(rawdata,ch,times,pre,post,channels,sampling_rate)).T
            probe_waveform[i,:,:] -= np.mean(probe_waveform[i,:,:])
        except:
            probe_waveform[i,:,:] = 0.#np.zeros((len(times),int(((pre+post)/1000.)*sampling_rate)))
    return probe_waveform.T

def load_phy_template(path,site_positions = option234_positions,**kwargs):
# load spike data that has been manually sorted with the phy-template GUI
# the site_positions should contain coordinates of the channels in probe space. for example, in um on the face of the probe
# returns a dictionary of 'good' units, each of which includes:
#	times: spike times, in seconds
#	template: template used for matching
#	ypos: y position on the probe, calculated from the template. requires an accurate site_positions. averages template from 100 spikes.
#	xpos: x position on the probe, calcualted from the template. requires an accurate site_positions. averages template from 100 spikes.
    clusters = np.load(open(os.path.join(path,'spike_clusters.npy'),'rb'))
    spikes = np.load(open(os.path.join(path,'spike_times.npy'),'rb'))
    spike_templates = np.load(open(os.path.join(path,'spike_templates.npy'),'rb'))
    templates = np.load(open(os.path.join(path,'templates.npy'),'rb'))
    cluster_id = [];
    #[cluster_id.append(row) for row in csv.reader(open(os.path.join(path,'cluster_group.tsv')))];
    if os.path.isfile(os.path.join(path,'cluster_group.tsv')):
        cluster_id = [row for row in csv.reader(open(os.path.join(path,'cluster_group.tsv')))];
    else:
        if os.path.isfile(os.path.join(path,'cluster_groups.csv')):
            cluster_id = [row for row in csv.reader(open(os.path.join(path,'cluster_groups.csv')))];
        else:
            print('could not find cluster groups csv or tsv')
            return None
    if 'sampling_rate' in kwargs.keys():
        samplingrate = kwargs['sampling_rate']
    else:
        samplingrate =30000.
        print('no sampling rate specified, using default of 30kHz')

    units = {}
    for i in np.arange(1,np.shape(cluster_id)[0]):
        if cluster_id[i][0].split('\t')[1] != 'garabge' :#:or cluster_id[i][0].split('\t')[1] == 'unsorted' :#if it is a 'good' cluster by manual sort

            unit = int(cluster_id[i][0].split('\t')[0])
            units[str(unit)] = {}

            #get the unit spike times
            units[str(unit)]['times'] = spikes[np.where(clusters==unit)]/samplingrate
            units[str(unit)]['times'] = units[str(unit)]['times'].flatten()

            #get the mean template used for this unit
            all_templates = spike_templates[np.where(clusters==unit)].flatten()
            n_templates_to_subsample = 100
            random_subsample_of_templates = templates[all_templates[np.array(np.random.rand(n_templates_to_subsample)*all_templates.shape[0]).astype(int)]]
            mean_template = np.mean(random_subsample_of_templates,axis=0)
            units[str(unit)]['template'] = mean_template

            #take a weighted average of the site_positions, where the weights is the absolute value of the template for that channel
            #this gets us the x and y positions of the unit on the probe.
            weights = np.zeros(site_positions.shape)
            for channel in range(site_positions.shape[0]):
                weights[channel,:]=np.trapz(np.abs(mean_template.T[channel,:]))
            weights = weights/np.max(weights)
            (xpos,ypos)=np.average(site_positions,axis=0,weights=weights)
            units[str(unit)]['waveform_weights'] = weights
            units[str(unit)]['xpos'] = xpos
            units[str(unit)]['ypos'] = ypos - site_positions[-1][1]

    return units    
##===============================================================================    



##===============================================================================
def ISIviolations(spikeTrain, refDur, minISI):
    #modified from cortex-lab/sortingQuality GitHub by nick s. 
    isis = np.diff(spikeTrain)
    nSpikes = len(spikeTrain)
    numViolations = sum(isis<refDur) 
    violationTime = 2*nSpikes*(refDur-minISI)
    totalRate = nSpikes/(spikeTrain[-1] - spikeTrain[0])
    violationRate = numViolations/violationTime
    fpRate = violationRate/totalRate
    if fpRate > 1.:
        fpRate = 1. # it is nonsense to have a rate > 1; a rate > 1 means the assumputions of this analysis are failing
    return fpRate, numViolations

def isiViolations(directory,time_limits=None,tr=.0015,tc=0):
    spike_clusters_path = os.path.join(directory,'spike_clusters.npy')
    spike_templates_path = os.path.join(directory,'spike_templates.npy')
    spike_times_path = os.path.join(directory,'spike_times.npy')
    params_path = os.path.join(directory,'params.py')
    
    if time_limits == None:
        time_limits=[0,1e7]

    print(' ')
    print('loading data for ISI computation...')
    if os.path.isfile(spike_clusters_path):
        spike_clusters = np.load(spike_clusters_path)
    else:
        spike_clusters = np.load(spike_templates_path)

    params = read_kilosort_params(params_path)
    spike_times = np.load(spike_times_path) / float(params['sample_rate'])

    print('computing ISI violations...')
    cluster_IDs = np.unique(spike_clusters)
    isiV = np.zeros(np.shape(cluster_IDs)[0])
    for i,cluster_ID in enumerate(cluster_IDs):
        all_spikes = spike_times[np.where(spike_clusters==cluster_ID)[0]].flatten()
        if all_spikes[-1] < time_limits[0] or all_spikes[0] > time_limits[1]:
            isiV[i] = np.nan
        else:
            spikes = all_spikes[np.where(all_spikes > time_limits[0])[0][0]:np.where(all_spikes < time_limits[1])[0][-1]]
            if len(spikes) > 1:
                (fp_rate, num_violations) = ISIviolations(spikes,tr,tc)
                #print fp_rate
                isiV[i] = fp_rate
                n_spikes = len(spikes)
                print('\rcluster '+str(cluster_ID)+': '+str(num_violations)+' violations ('+str(n_spikes)+' spikes), '+str(fp_rate)+' estimated FP rate',)
            else:
                isiV[i] = np.nan
    
    return cluster_IDs,isiV
    
def cluster_signalToNoise(directory,time_limits=None,filename='experiment1_100-0_0.dat',sigma=5.,number_to_average=250,plots=False,channels='all'):
    spike_clusters_path = os.path.join(directory,'spike_clusters.npy')
    spike_templates_path = os.path.join(directory,'spike_templates.npy')
    spike_times_path = os.path.join(directory,'spike_times.npy')
    params_path = os.path.join(directory,'params.py')
    
    if time_limits == None:
        time_limits=[0,1e7]

    print(' ')
    print('loading data for s/n computation...')
    if os.path.isfile(spike_clusters_path):
        spike_clusters = np.load(spike_clusters_path)
    else:
        spike_clusters = np.load(spike_templates_path)
    try:
        rawdata = np.memmap(os.path.join(directory,filename), dtype=np.int16, mode='r')
    except:
        print('could not load spike data. is the filename correct? (default: experiment1_100-0_0.dat)')
        return None
    spike_templates = np.load(spike_templates_path)
    templates = np.load(os.path.join(directory,'templates.npy'))
    
    params = read_kilosort_params(params_path)
    spike_times = np.load(spike_times_path) / float(params['sample_rate'])
    if channels == 'all':
        channels = get_channel_count(directory)
    
    site_positions=option234_positions[np.linspace(0,channels-1,channels).astype(int)]
    data = load_phy_template(directory,site_positions)
   
    print('computing Quian Quiroga signal/noise ratio...')
    cluster_IDs = np.unique(spike_clusters)
    sn_peak = np.zeros(np.shape(cluster_IDs)[0])
    sn_mean = np.zeros(np.shape(cluster_IDs)[0])
    #spike_amplitudes = []
    sn_all=[]
    for i,cluster_ID in enumerate(cluster_IDs):
        print('\rcluster '+str(cluster_ID)+': '+str(i)+'/'+str(len(cluster_IDs)),)
        #peak_y_channel = np.where(data[clusterID]['waveform_weights'] == np.max(data[clusterID]['waveform_weights']))[0][0]
        all_spikes = spike_times[np.where(spike_clusters==cluster_ID)[0]].flatten()
        if all_spikes[-1] < time_limits[0] or all_spikes[0] > time_limits[1] or str(cluster_ID) not in data.keys():
            sn_peak[i] = np.nan
            sn_mean[i] = np.nan
            sn_all.append([np.nan])
        else:
            spikes = all_spikes[np.where(all_spikes > time_limits[0])[0][0]:np.where(all_spikes < time_limits[1])[0][-1]]
            channels_with_template = [ch for ch,t in enumerate(data[str(cluster_ID)]['waveform_weights']) if np.max(np.abs(t))>0.1] # <----- get the channels that have a template > 20uV
            if len(spikes) > number_to_average:
                sub_times = np.random.choice(spikes,number_to_average,replace=False)
            else:
                sub_times = spikes
            random_times = np.random.rand(number_to_average) * (np.max(spike_times)-np.min(spike_times)) + np.min(spike_times)
            s_ns = []
            if plots:
                plt.figure()
            for channel in channels_with_template:
                ws = load_waveforms(os.path.join(directory,filename),channel,sub_times,channels=channels)
                rs = load_waveforms(os.path.join(directory,filename),channel,random_times,channels=channels)
                
                mean = np.mean(ws,axis=0)       

                # if time_limits[1]==1e7:
                #     cunk = rawdata[int(650*int(params['sample_rate'].strip('.'))*channels):int(660*int(params['sample_rate'].strip('.'))*channels)][channel::channels] 
                # else:
                #     cunk = rawdata[int(time_limits[0]*int(params['sample_rate'].strip('.'))*channels):int(time_limits[1]*int(params['sample_rate'].strip('.'))*channels)][channel::channels] 
                # channel_chunk = filtr(cunk,300,6000,float(params['sample_rate']),3) * 0.195 # <-- re-filter and put into uV
                
                #noise = sigma * np.median(np.abs(channel_chunk)/0.6725)
                noise = sigma * np.median(np.abs(np.array(rs))/0.6725) # <--- as defined by Quian Quiroga 2004
                signal = np.max(np.abs(mean)) # <-- convert to uV
                # if signal/noise > np.max(s_ns):
                #     amplitudes = [np.max(np.abs(w)) for w in np.array(waveforms).T]
                s_ns.extend([signal/noise])
                if plots:
                    plt.plot(mean)
            if plots:
                plt.title(str(np.max(s_ns)))
    
            sn_peak[i] = np.max(s_ns)
            sn_mean[i] = np.mean(s_ns)
            sn_all.append(s_ns)
            #spike_amplitudes.append(amplitudes)
            print('\rcluster '+str(cluster_ID)+': '+str(np.max(s_ns))+' sn',)
    return cluster_IDs, sn_peak, sn_mean, sn_all# spike_amplitudes,sn_all

def masked_cluster_quality(directory,time_limits=None,n_fet=3,minimum_number_of_spikes=10):
    pc_features_path = os.path.join(directory,'pc_features.npy')
    pc_features_ind_path = os.path.join(directory,'pc_feature_ind.npy')
    spike_clusters_path = os.path.join(directory,'spike_clusters.npy')
    spike_templates_path = os.path.join(directory,'spike_templates.npy')
    spike_times_path = os.path.join(directory,'spike_times.npy')
    params_path = os.path.join(directory,'params.py')
    params = read_kilosort_params(params_path)
    
    try:
        pc_features = np.load(pc_features_path)
    except:
        print('loading PC features failed.')
        return None
    pc_feature_ind = np.load(pc_features_ind_path)

    if os.path.isfile(spike_clusters_path):
        print('building features matrix from clusters / templates')
        spike_clusters = np.load(spike_clusters_path)
        spike_templates = np.load(spike_templates_path)
        spike_times = np.load(spike_times_path) / float(params['sample_rate'])
        
        cluster_IDs = np.unique(spike_clusters)
        n_clusters = len(cluster_IDs)
        n_spikes = len(spike_clusters)
        n_fet_per_chan = np.shape(pc_features)[1]
        n_templates = np.shape(pc_feature_ind)[0]
        new_fet = np.zeros((n_spikes,n_fet_per_chan,n_fet))
        new_fet_inds = np.zeros((n_clusters,n_fet))
        
        if time_limits == None:
            time_limits=[0,1e7]

    else:
        print('spike_clusters do not exist, using spike_templates instead')
    
    print('computing cluster qualities...')
    (cluster_IDs,unit_quality,contamination_rate,flda) = masked_cluster_quality_sparse(spike_clusters,pc_features,pc_feature_ind,spike_times,spike_templates,time_limits)
    return cluster_IDs,unit_quality,contamination_rate, flda


def masked_cluster_quality_sparse(spike_clusters,pc_features,pc_feature_ind,spike_times,spike_templates,time_limits=None,n_fet=3,fet_n_chans=6):
    fet_N = np.shape(pc_features)[1] * fet_n_chans
    N = len(spike_clusters)
    cluster_IDs = np.unique(spike_clusters)
    unit_quality = np.zeros(len(cluster_IDs))
    contamination_rate = np.zeros(len(cluster_IDs))
    flda = np.zeros(len(cluster_IDs))
    print('number of clusters: '+str(np.shape(unit_quality)[0]))
    if time_limits == None:
        time_limits=[0,1e7]
        
    for i,cluster_ID in enumerate(cluster_IDs):
        all_spikes = spike_times[np.where(spike_clusters==cluster_ID)[0]].flatten()
        if all_spikes[-1] < time_limits[0] or all_spikes[0] > time_limits[1]:
            unit_quality[i] = 0;
            contamination_rate[i] =np.nan;
        else:
            these_spikes = np.where(spike_clusters==cluster_ID)[0][np.where(all_spikes > time_limits[0])[0][0]:np.where(all_spikes < time_limits[1])[0][-1]+1]
            n_spikes_in_cluster = len(these_spikes)
            if n_spikes_in_cluster < fet_n_chans or n_spikes_in_cluster > N/2.:
                unit_quality[i] = 0;
                contamination_rate[i] =np.nan;
            else:
                fet_this_cluster = pc_features[these_spikes,:n_fet,:fet_n_chans]#.reshape((len(these_spikes),-1))
                #print 'this cluster PCs: '+str(np.shape(fet_this_cluster))
                
                these_templates = spike_templates[these_spikes]
                #count the templates in this unit and their frequency of occurence:
                (included_templates,instances) = count_unique(these_templates)
                #use the template that occurs most frequently:
                this_template = included_templates[np.where(instances==np.max(instances))[0][0]]
                
                this_cluster_chans = pc_feature_ind[this_template,:fet_n_chans]
                other_clusters_IDs = []
                fet_other_clusters = []
                for ii,cluster_2 in enumerate(cluster_IDs):
                    try:
                        if cluster_2 != cluster_ID:
                            all_spikes_2 = spike_times[np.where(spike_clusters==cluster_2)[0]].flatten()
                            these_spikes_2 = np.where(spike_clusters==cluster_2)[0][np.where(all_spikes_2 > time_limits[0])[0][0]:np.where(all_spikes_2 < time_limits[1])[0][-1]+1]
                            these_templates_2 = spike_templates[these_spikes_2]
                            #count the templates in this unit and their frequency of occurence:
                            (included_templates_2,instances_2) = count_unique(these_templates_2)
                            #use the template that occurs most frequently:
                            this_template_2 = included_templates_2[np.where(instances_2==np.max(instances_2))[0][0]]
                            #get the channels that have signal in this template
                            cluster_2_chans = pc_feature_ind[this_template_2]
                                    
                            if np.any(np.in1d(this_cluster_chans,cluster_2_chans)):  #<---continue only if the second cluster has an overlapping channel with the main cluster  
                                #spikes_other_clusters.extend(these_spikes_2)
                                fet_cluster2 = np.zeros((np.shape(these_spikes_2)[0],n_fet,fet_n_chans)) #<---make an empty matrix with dimensions: [number of spikes,number of features, number of channels]
                                #get the pca values, for each channel and each feature, that overlap with the main cluster in question
                                for iii,ch in enumerate(this_cluster_chans):
                                    if np.in1d(this_cluster_chans,cluster_2_chans)[iii]:
                                        fet_cluster2[:,:,iii] = pc_features[these_spikes_2,:n_fet,np.where(ch==cluster_2_chans)[0][0]]
                                fet_other_clusters.extend(fet_cluster2)
                    except:
                        pass
                            
                #reshape to the arrays to dimensions: [number of spikes, (number of features * number of channels)]
                try:
                    fet_other_clusters = np.array(fet_other_clusters).reshape((np.shape(fet_other_clusters)[0],-1))
                    fet_this_cluster = fet_this_cluster.reshape((len(these_spikes),-1))
                except ValueError:
                    pass

                #pass the features to core to calculate isolation distance and mahalanobis contamination
                unit_quality[i],contamination_rate[i] = masked_cluster_quality_core(fet_this_cluster,fet_other_clusters,plots=False)
                
                #pass the features to flda to calculate the d prime of this cluster from all other spikes, using Fisher's LDA
                try:
                    flda[i] = masked_cluster_quality_flda(fet_this_cluster,fet_other_clusters,plots=False)
                except ValueError:
                    pass
                #pass spike times to neighbor similarity to measure the distance between 
                #neighbor_similarity[i],spike_similarity = masked_waveform_similarity(these_spikes,spikes_other_clusters,number_to_average=250,metric='euclidean')
                
                #plans for the future:
                #l_ratio[i] = masked_cluster_quality_lratio(fet_this_cluster,fet_other_clusters,plots=False)

            
        print('\rcluster '+str(cluster_ID)+': # spikes:'+str(np.shape(all_spikes)[0])+'  iso. distance:'+str(unit_quality[i])+'  contamination:'+str(contamination_rate[i])+'  '+str(i+1)+'/'+str(np.shape(unit_quality)[0]),)
    return cluster_IDs,unit_quality,contamination_rate,flda

# def masked_waveform_similarity(directory,number_to_average=250,metric='euclidean'):
#     params_path = os.path.join(directory,'params.py')
#     filename = read_kilosort_params(params_path)['dat_path'][1:-1]
#     
#     if len(spikes) > number_to_average:
#         sub_times = np.random.choice(spikes,number_to_average,replace=False)
#     else:
#         sub_times = spikes
#         
#     random_times = np.random.rand(number_to_average) * (np.max(spike_times)-np.min(spike_times)) + np.min(spike_times)
# 
#     for channel in channels_with_template:
#         ws = load_waveforms(os.path.join(directory,filename),channel,sub_times,channels=channels)
#         
#     return neighbor_similarity_dprime,spike_by_spike_similarity

def masked_cluster_quality_flda(fet_this_cluster,fet_other_clusters,plots=False):
    #based on documentation here: http://sebastianraschka.com/Articles/2014_python_lda.html
    flda =LDA(n_components=1)
    X = np.concatenate((fet_this_cluster,
                        fet_other_clusters))
    y = np.concatenate((np.zeros(np.shape(fet_this_cluster)[0]),
                        np.ones(np.shape(fet_other_clusters)[0])))
    X_flda = flda.fit_transform(X, y)
    
    flda_this_cluster  = X_flda[:np.shape(fet_this_cluster)[0]]
    flda_other_cluster = X_flda[np.shape(fet_this_cluster)[0]:]
    
    dprime = (np.mean(flda_this_cluster) - np.mean(flda_other_cluster))/np.sqrt(0.5*(np.std(flda_this_cluster)**2+np.std(flda_other_cluster)**2))
    return dprime

def masked_cluster_quality_core(fet_this_cluster,fet_other_clusters,point_limit=20000000,plots=False):
    n = np.shape(fet_this_cluster)[0]
    n_other = np.shape(fet_other_clusters)[0]
    n_fet = np.shape(fet_this_cluster)[1]

    if n_other > n and n > n_fet:
        if n > point_limit:
            random_indices = np.random.choice(n,point_limit,replace=False)
            fet_this_cluster = fet_this_cluster[random_indices,:]
        if n_other > point_limit:
            random_indices = np.random.choice(n_other,point_limit,replace=False)
            fet_other_clusters = fet_other_clusters[random_indices,:]

        md = np.sort(cdist(fet_this_cluster.mean(0).reshape(1,fet_this_cluster.shape[1]),
                            fet_other_clusters,
                            'mahalanobis')[0])
        md_self = np.sort(cdist(fet_this_cluster.mean(0).reshape(1,fet_this_cluster.shape[1]),
                                 fet_this_cluster,
                                 'mahalanobis')[0])
        #print fet_this_cluster.mean(0).reshape(1,fet_this_cluster.shape[1])
        if plots:
            plt.figure()
            plt.plot(fet_this_cluster[:,0],fet_this_cluster[:,1],'r.')
            plt.plot(fet_other_clusters[:n,0],fet_other_clusters[:n,1],'b.')
            plt.plot(fet_this_cluster.mean(0).reshape(1,fet_this_cluster.shape[1])[0][0],fet_this_cluster.mean(0).reshape(1,fet_this_cluster.shape[1])[0][1],'*',color='#ffcccc',ms=12)
            #
            plt.figure()
            plt.hist(md[:n],bins=100,range=(0,10),color='b')
            plt.hist(md_self,bins=100,range=(0,10),color='r')
            plt.title('iso: '+str(np.max(md[:n])))
        unit_quality = np.max(md[:n])
        contamination_rate = 1 - (tipping_point(md_self,md) / float(len(md_self)))
    else:
        unit_quality = 0
        contamination_rate = np.nan
    return unit_quality, contamination_rate



    
def tipping_point(x,y):
# Input: x, y  are sorted ascending arrays of positive numbers
# Output: minimal pos s.t. sum(x > x(pos)) <= sum(y < x(pos))

#original:
# algorithm here is to sort x and y together, and determine the indices of
# x in this sorted list (call this xInds). Then, xInds-(1:length(xInds))
# will be the number of y's that are less than that value of x.

# translated from matlab (by nick steinmetz) to python:
# algorithm here is to sort x and y together, and determine how many y are less than x 
# in the first len(x) instances of the sorted together array.


    #pos = [ind for ind,together in enumerate(np.sort(np.array(zip(x,y)).flatten())) if np.any(together==x)][:len(x)][-1]-len(x)
    #pos = np.where(np.array(np.sort(np.concatenate((x,y)).flatten()))[:len(x)] == np.sort(x)[-1])[0][0] - len(x)
    #print str(np.where(np.array(np.sort(np.array(zip(x,y)).flatten()))== np.sort(x)[-1])[0][0])+' furthest self: '+str(x[-1])+'    closest other:'+str(y[0])
    
    pos =  sum(np.in1d(np.array(np.sort(np.concatenate((x,y)).flatten()))[:len(x)],np.sort(x)))

    return pos






##===============================================================================
##plotting functions
##===============================================================================

def placeAxesOnGrid(fig,dim=[1,1],xspan=[0,1],yspan=[0,1],wspace=None,hspace=None,):
    '''
    Takes a figure with a gridspec defined and places an array of sub-axes on a portion of the gridspec
    
    Takes as arguments:
        fig: figure handle - required
        dim: number of rows and columns in the subaxes - defaults to 1x1
        xspan: fraction of figure that the subaxes subtends in the x-direction (0 = left edge, 1 = right edge)
        yspan: fraction of figure that the subaxes subtends in the y-direction (0 = top edge, 1 = bottom edge)
        wspace and hspace: white space between subaxes in vertical and horizontal directions, respectively
        
    returns:
        subaxes handles
        
    written by doug ollerenshaw
    '''
    import matplotlib.gridspec as gridspec

    outer_grid = gridspec.GridSpec(100,100)
    inner_grid = gridspec.GridSpecFromSubplotSpec(dim[0],dim[1],
                                                  subplot_spec=outer_grid[int(100*yspan[0]):int(100*yspan[1]),int(100*xspan[0]):int(100*xspan[1])],
                                                  wspace=wspace, hspace=hspace)
    

    #NOTE: A cleaner way to do this is with list comprehension:
    # inner_ax = [[0 for ii in range(dim[1])] for ii in range(dim[0])]
    inner_ax = dim[0]*[dim[1]*[fig]] #filling the list with figure objects prevents an error when it they are later replaced by axis handles
    inner_ax = np.array(inner_ax)
    idx = 0
    for row in range(dim[0]):
        for col in range(dim[1]):
            inner_ax[row][col] = plt.Subplot(fig, inner_grid[idx])
            fig.add_subplot(inner_ax[row,col])
            idx += 1

    inner_ax = np.array(inner_ax).squeeze().tolist() #remove redundant dimension
    return inner_ax
def cleanAxes(ax,bottomLabels=False,leftLabels=False,rightLabels=False,topLabels=False,total=False):
    ax.tick_params(axis='both',labelsize=10)
    ax.spines['top'].set_visible(False);
    ax.yaxis.set_ticks_position('left');
    ax.spines['right'].set_visible(False);
    ax.xaxis.set_ticks_position('bottom')
    if not bottomLabels or topLabels:
        ax.set_xticklabels([])
    if not leftLabels or rightLabels:
        ax.set_yticklabels([])
    if rightLabels:
        ax.spines['right'].set_visible(True);
        ax.spines['left'].set_visible(False);
        ax.yaxis.set_ticks_position('right');
    if total:
        ax.set_frame_on(False);
        ax.set_xticklabels('',visible=False);
        ax.set_xticks([]);
        ax.set_yticklabels('',visible=False);
        ax.set_yticks([])
def psth_line(times,triggers,pre=0.5,timeDomain=False,post=1,binsize=0.05,ymax=75,yoffset=0,output='fig',name='',color='#00cc00',linewidth=0.5,axes=None,labels=True,sparse=False,labelsize=18,axis_labelsize=20,error='',alpha=0.5,**kwargs):
    post = post + 1
    peris=[]#np.zeros(len(triggers),len(times))
    p=[]
    if timeDomain:
        samplingRate = 1.0
    else:
        samplingRate = samplingRate
        
    times = np.array(times).astype(float) / samplingRate + pre
    triggers = np.array(triggers).astype(float) / samplingRate

    numbins = (post+pre) / binsize 
    bytrial = np.zeros((len(triggers),numbins))
    for i,t in enumerate(triggers):
        
        if len(np.where(times >= t - pre)[0]) > 0 and len(np.where(times >= t + post)[0]) > 0:
            start = np.where(times >= t - pre)[0][0]
            end = np.where(times >= t + post)[0][0]
            for trial_spike in times[start:end-1]:
                if float(trial_spike-t)/float(binsize) < float(numbins):
                    bytrial[i][(trial_spike-t)/binsize-1] +=1   
        else:
        	 pass
             #bytrial[i][:]=0
        #print 'start: ' + str(start)+'   end: ' + str(end)

    variance = np.std(bytrial,axis=0)/binsize/np.sqrt((len(triggers)))
    hist = np.mean(bytrial,axis=0)/binsize
    edges = np.linspace(-pre,post,numbins)

    if output == 'fig':
        if error == 'shaded':
            if 'shade_color' in kwargs.keys():
                shade_color=kwargs['shade_color']
            else:
                shade_color=color    
            if axes == None:
                plt.figure()
                axes=plt.gca()
            plt.locator_params(axis='y',nbins=4)
            upper = hist+variance
            lower = hist-variance
            axes.fill_between(edges[2:-1],upper[2:-1]+yoffset,hist[2:-1]+yoffset,alpha=alpha,color='white',facecolor=shade_color)
            axes.fill_between(edges[2:-1],hist[2:-1]+yoffset,lower[2:-1]+yoffset,alpha=alpha,color='white',facecolor=shade_color)
            axes.plot(edges[2:-1],hist[2:-1]+yoffset,color=color,linewidth=linewidth)
            axes.set_xlim(-pre,post-1)
            axes.set_ylim(0,ymax);
            if sparse:
                axes.set_xticklabels([])
                axes.set_yticklabels([])
            else:
                if labels:
                    axes.set_xlabel(r'$time \/ [s]$',fontsize=axis_labelsize)
                    axes.set_ylabel(r'$firing \/ rate \/ [Hz]$',fontsize=axis_labelsize)
                    axes.tick_params(axis='both',labelsize=labelsize)
            axes.spines['top'].set_visible(False);axes.yaxis.set_ticks_position('left')
            axes.spines['right'].set_visible(False);axes.xaxis.set_ticks_position('bottom')   
            axes.set_title(name,y=0.5)
            return axes 
        else:
            if axes == None:
                plt.figure()
                axes=plt.gca()
            f=axes.errorbar(edges,hist,yerr=variance,color=color)
            axes.set_xlim(-pre,post - 1)
            if ymax=='auto':
                pass
            else:
                axes.set_ylim(0,ymax)
            if sparse:
                axes.set_xticklabels([])
                axes.set_yticklabels([])
            else:
                if labels:
                    axes.set_xlabel(r'$time \/ [s]$',fontsize=axis_labelsize)
                    axes.set_ylabel(r'$firing \/ rate \/ [Hz]$',fontsize=axis_labelsize)
                    axes.tick_params(axis='both',labelsize=labelsize)
            axes.spines['top'].set_visible(False);axes.yaxis.set_ticks_position('left')
            axes.spines['right'].set_visible(False);axes.xaxis.set_ticks_position('bottom')   
            axes.set_title(name)
            return axes
    if output == 'hist':
        return (hist,edges)    
    if output == 'p':
        return (edges,hist,variance)
        

def project_linear_quality(data,labels,plots=False):
    flda =LDA(n_components=1)
    if np.shape(data)[0] == np.shape(labels)[0]:
        X_flda = flda.fit_transform(data, labels)
        return X_flda
    else:
        print('dimensions of data do not match labels')
        return None

def get_spike_amplitudes(datapath,channel,times,pre=0.5,post=1.5,channels=384,sampling_rate=30000):
    pre = pre * .001
    post = post * .001
    mm = np.memmap(datapath, dtype=np.int16, mode='r')
    amplitudes = []
    channel = int(channel)
    for i in times:
        start = int((i - pre) * sampling_rate) * int(channels)
        temp=mm[start:start+int((pre+post)*sampling_rate*channels)][channel::channels]
        amplitudes.extend([np.max(np.abs(temp-temp[0]))])
    return amplitudes

def get_PCs(datapath,channel,times,PC=1,pre=0.5,post=1.5,channels=384,sampling_rate=30000):
    pre = pre * .001
    post = post * .001
    mm = np.memmap(datapath, dtype=np.int16, mode='r')
    PCs = []
    for i in times:
        start = int((i - pre) * sampling_rate) * channels
        amplitudes.extend([np.max(np.abs(mm[start:start+(pre+post)*sampling_rate*channels][channel::channels] - mm[start:start+(pre+post)*sampling_rate*channels][channel::channels][0]))])
    return PCs

# %%

def neuron_fig(clusterID,df,sortpath,filename='experiment1_102-0_0.dat',time_limits=None,timeplot_binsize=60.,neighbor_colors=["#67572e","#50a874","#ff4d4d"]):

    channels = 175 #get_channel_count(sortpath)
    site_positions=option234_positions[np.linspace(0,channels-1,channels).astype(int)]
    data = load_phy_template(sortpath,site_positions)
    
    cluster_ID = int(clusterID)
    cluster_IDs = np.load(os.path.join(sortpath,'spike_clusters.npy'))
    pc_data = np.load(os.path.join(sortpath,'pc_features.npy'))
    pc_ind_data = np.load(os.path.join(sortpath,'pc_feature_ind.npy'))
    params_path = os.path.join(sortpath,'params.py')
    params = read_kilosort_params(params_path)
    spike_times_data = np.load(os.path.join(sortpath,'spike_times.npy'))/ float(params['sample_rate'])
    spike_templates = np.load(os.path.join(sortpath,'spike_templates.npy'))
    
    datapath=os.path.join(sortpath,filename)
    #df_line = df[df.clusterID==int(clusterID)]
    
    fig = plt.figure(figsize=(11,8.5))
    ax_text = placeAxesOnGrid(fig,xspan=[0.0,0.4],yspan=[0,0.1])
    ax_position = placeAxesOnGrid(fig,xspan=[0,0.1],yspan=[0.12,1.0])
    ax_waveform = placeAxesOnGrid(fig,xspan=[0.2,0.45],yspan=[0.12,0.65])
    ax_time = placeAxesOnGrid(fig,xspan=[0.2,1.0],yspan=[0.82,1.0])
    ax_PCs = placeAxesOnGrid(fig,xspan=[0.5,0.8],yspan=[0,0.35])
    ax_ACG = placeAxesOnGrid(fig,dim=[1,2],xspan=[0.55,1.0],yspan=[0.5,0.7])
    #ax_neighbor_waveform_1 = placeAxesOnGrid(fig,dim=[1,1],xspan=[0.82,1.0],yspan=[0,0.13])
    #ax_neighbor_waveform_2 = placeAxesOnGrid(fig,dim=[1,1],xspan=[0.82,1.0],yspan=[0.13,0.26])
    #ax_neighbor_waveform_3 = placeAxesOnGrid(fig,dim=[1,1],xspan=[0.82,1.0],yspan=[0.26,0.39])
    ax_neighbor_waveforms=placeAxesOnGrid(fig,dim=[1,1],xspan=[0.82,1.0],yspan=[0.0,0.39])#[ax_neighbor_waveform_1,ax_neighbor_waveform_2,ax_neighbor_waveform_3]
    ax_CCGs_1 = placeAxesOnGrid(fig,dim=[1,1],xspan=[0.53,0.68],yspan=[0.36,0.48])
    ax_CCGs_2 = placeAxesOnGrid(fig,dim=[1,1],xspan=[0.7,.85],yspan=[0.36,0.48])
    ax_CCGs_3 = placeAxesOnGrid(fig,dim=[1,1],xspan=[0.86,1.0],yspan=[0.36,0.48])
    ax_CCGs = [ax_CCGs_1,ax_CCGs_2,ax_CCGs_3]
    
    #position plot
    ax_position.imshow(data[clusterID]['waveform_weights'][::4],extent=(site_positions[:,0][::4][0],site_positions[:,0][::4][0]+16,site_positions[:,1][::4][0],site_positions[:,1][::4][-1]),cmap=plt.cm.gray_r,clim=(0,0.5),interpolation='none')
    ax_position.imshow(data[clusterID]['waveform_weights'][1::4],extent=(site_positions[:,0][1::4][0],site_positions[:,0][1::4][0]+16,site_positions[:,1][1::4][0],site_positions[:,1][1::4][-1]),cmap=plt.cm.gray_r,clim=(0,0.5),interpolation='none')
    ax_position.imshow(data[clusterID]['waveform_weights'][2::4],extent=(site_positions[:,0][2::4][0],site_positions[:,0][2::4][0]+16,site_positions[:,1][2::4][0],site_positions[:,1][2::4][-1]),cmap=plt.cm.gray_r,clim=(0,0.5),interpolation='none')
    ax_position.imshow(data[clusterID]['waveform_weights'][3::4],extent=(site_positions[:,0][3::4][0],site_positions[:,0][3::4][0]+16,site_positions[:,1][3::4][0],site_positions[:,1][3::4][-1]),cmap=plt.cm.gray_r,clim=(0,0.5),interpolation='none')
    ax_position.set_aspect(0.1)
    ax_position.set_ylim(3840,0)
    ax_position.set_xlim(70,0)
    cleanAxes(ax_position)
    ax_position.set_title('neuron position')
    
    #time limits
    if time_limits == None:
        time_limits=[0,1e7]
    all_spikes = data[clusterID]['times']
    spike_times = all_spikes[np.where(all_spikes > time_limits[0])[0][0]:np.where(all_spikes < time_limits[1])[0][-1]]
    #print len(all_spikes)
    #print all_spikes[0]
    #print all_spikes[-1]
    ##print np.where(all_spikes > time_limits[0])[0][0]
    #print np.where(all_spikes < time_limits[1])[0][-1]
    these_spikes = np.where(cluster_IDs==cluster_ID)[0][np.where(all_spikes > time_limits[0])[0][0]:np.where(all_spikes < time_limits[1])[0][-1]]
    #spike_times = spike_times_data[these_spikes]

    #for PC and CCG display, find close by clusters calculate 
    #PC plot
    number_of_spikes_to_plot = 5000
    these_templates=spike_templates[np.where(cluster_IDs==cluster_ID)[0]]
    (included_templates,instances) = count_unique(these_templates)
    this_template = included_templates[int(np.where(instances==np.max(instances))[0])]
    ch1 = pc_ind_data[this_template][0]
    ch2 = pc_ind_data[this_template][1]
    ax_PCs.plot(pc_data[these_spikes][:number_of_spikes_to_plot,0,0],
             pc_data[these_spikes][:number_of_spikes_to_plot,0,1],'bo',ms=1.5,markeredgewidth=0)
    
    nearby_trio = [0,0,0];
    nearby_euclids = [10000,10000,10000];
    nearby_times=[]
    for other_cluster in data.keys():
        if other_cluster != clusterID:
            if (np.abs(data[clusterID]['ypos']-data[other_cluster]['ypos']) + np.abs(data[clusterID]['xpos']-data[other_cluster]['xpos'])) < nearby_euclids[-1]:
                rank = np.where((np.abs(data[clusterID]['ypos']-data[other_cluster]['ypos']) + np.abs(data[clusterID]['xpos']-data[other_cluster]['xpos'])) < nearby_euclids)[0][0]
                nearby_euclids[rank] = (np.abs(data[clusterID]['ypos']-data[other_cluster]['ypos']) + np.abs(data[clusterID]['xpos']-data[other_cluster]['xpos']))
                nearby_trio[rank]=other_cluster
    print(nearby_trio)
    for ii,neighbor in enumerate(nearby_trio):
        all_spikes_neighbor = data[neighbor]['times']
        indices_neighbor= np.where(cluster_IDs==int(neighbor))[0][np.where(all_spikes_neighbor > time_limits[0])[0][0]:np.where(all_spikes_neighbor < time_limits[1])[0][-1]]
        neighbor_templates=spike_templates[indices_neighbor]
        neighbor_spike_times = all_spikes_neighbor[np.where(all_spikes_neighbor > time_limits[0])[0][0]:np.where(all_spikes_neighbor < time_limits[1])[0][-1]]
        nearby_times.append(neighbor_spike_times)
        (included_templates,instances) = count_unique(neighbor_templates)
        this_template = included_templates[int(np.where(instances==np.max(instances))[0])]
        ch1_index = np.where(pc_ind_data[this_template] == ch1)[0]
        ch2_index = np.where(pc_ind_data[this_template] == ch2)[0]
        if ch2_index.size != 0 and ch1_index.size != 0:
            ax_PCs.plot(pc_data[indices_neighbor][:number_of_spikes_to_plot,0,np.where(pc_ind_data[this_template] == ch1)[0][0]],
                     pc_data[indices_neighbor][:number_of_spikes_to_plot,0,np.where(pc_ind_data[this_template] == ch2)[0][0]],
                     'o',color=neighbor_colors[ii],ms=1,markeredgewidth=0,alpha=0.8)
            
            all_diffs = []
            for spike_time in spike_times:
                try:
                    neighbor_start = np.where(neighbor_spike_times < spike_time - 0.5)[0][-1]
                except:
                    neighbor_start = 0
                try:
                    neighbor_end = np.where(neighbor_spike_times > spike_time + 0.5)[0][0]
                except:
                    neighbor_end = -1
                neighbor_chunk = neighbor_spike_times[neighbor_start:neighbor_end]
                #print '\r'+str(spike_time)+' '+str(neighbor_start)+' - '+str(neighbor_end),
                all_diffs.extend(neighbor_chunk - spike_time)
            all_diffs=np.array(all_diffs).flatten()
            hist,edges = np.histogram(all_diffs,bins=np.linspace(-0.2,0.2,400))
            ax_CCGs[ii].plot(edges[:-1],hist,drawstyle='steps',color=neighbor_colors[ii])
            ax_CCGs[ii].set_xlim(-.02,.02)
            ax_CCGs[ii].xaxis.set_ticks([-0.075,0.0,0.075])
            ax_CCGs[ii].axvline(0.0,ymax=0.1,ls='--',color='#ff8080')
        cleanAxes(ax_CCGs[ii],total=True)
    cleanAxes(ax_neighbor_waveforms,total=True)
    cleanAxes(ax_PCs)    
    ax_PCs.set_title('PC features')
    
    #waveform plot
    cleanAxes(ax_waveform,total=True)
    ax_waveform.set_title('waveform')
    channel_offset = 0 # should not be 0 if not starting from tip.
    peak_y_channel = np.where(data[clusterID]['waveform_weights'] == np.max(data[clusterID]['waveform_weights']))[0][0]
    times = np.random.choice(spike_times,100,replace=False)
    random_times = np.random.rand(100) * (np.max(spike_times)-np.min(spike_times)) + np.min(spike_times)
    xoffs=[0,100,50,150,0,100,50,150,0,100,50,150,0,100,50,150,0,100][::-1]
    yoff=100
    signal = 0;
    for ii,channel in enumerate(np.linspace(peak_y_channel-8,peak_y_channel+10,18)):
        if channel > 0:
            ws = load_waveforms(datapath,channel,times,channels=channels)
            ws_bkd = load_waveforms(datapath,channel,random_times,channels=channels)
            x_range = np.linspace(xoffs[ii],xoffs[ii]+60,60)
            for i in range(np.shape(ws)[0]):
                if np.shape(x_range)[0] == np.shape(ws[i][:]-yoff*(ii/2))[0]:
                    ax_waveform.plot(x_range,ws[i][:]-yoff*(ii/2),alpha=0.05,color='#0066ff')
                    ax_waveform.plot(x_range,ws_bkd[i][:]-yoff*(ii/2),alpha=0.05,color='#c8c8c8')
            if np.shape(x_range)[0] == np.shape(np.mean(ws,axis=0)-yoff*(ii/2))[0]:
                if np.max(np.abs(np.mean(ws,axis=0)-yoff*(ii/2))) > signal:
                    signal = np.max(np.abs(np.mean(ws,axis=0)-yoff*(ii/2)))
                    noise = np.mean(ws_bkd) + np.std(ws_bkd)* 4.
                ax_waveform.plot(x_range,np.mean(ws,axis=0)-yoff*(ii/2),color='#0066ff')
            if ii > 3 and ii < 14:
                for nn,axis in enumerate(nearby_trio):
                        if np.shape(x_range)[0] == np.shape(np.mean(ws,axis=0)-yoff*(ii/2))[0]:
                            neighbor_ws = load_waveforms(datapath,channel,np.random.choice(nearby_times[nn],100,replace=False),channels=channels)
                            ax_neighbor_waveforms.plot(x_range,np.mean(ws,axis=0)-(yoff/3.)*(ii/2),color='#0066ff')
                            ax_neighbor_waveforms.plot(x_range,np.mean(neighbor_ws,axis=0)-(yoff/3.)*(ii/2),color=neighbor_colors[nn],alpha=0.8)

    #time plot
    hist,edges = np.histogram(spike_times,bins=int(np.ceil(spike_times[-1] / timeplot_binsize)))
    ax_time.plot(edges[1:],hist/float(timeplot_binsize),drawstyle='steps')  
    ax_time.set_xlabel('time (sec)')
    ax_time.set_ylabel('firing rate (Hz)')
    ax_time.set_title('firing rate over time')
    max_spikes_to_plot = 2000
    if len(spike_times) > max_spikes_to_plot:
        times = np.random.choice(spike_times,max_spikes_to_plot,replace=False)
    else:
        times = spike_times
    amps = get_spike_amplitudes(datapath,peak_y_channel,times,channels=channels)
    ax_time_r = ax_time.twinx()
    ax_time_r.plot(times,amps,'o',markersize=2,alpha=0.3) 
    ax_time_r.set_ylabel('amplitude')
    ax_time_r.set_ylim(0,np.max(amps))
    cleanAxes(ax_time,leftLabels=True,bottomLabels=True)
    cleanAxes(ax_time_r,rightLabels=True,bottomLabels=True)
    
    #for ACG display, calculate ISIs.
    isis = np.diff(spike_times)
    hist,edges = np.histogram(np.concatenate((isis,isis*-1)),bins=np.linspace(-0.5,0.5,100))
    ax_ACG[0].plot(edges[:-1],hist,drawstyle='steps')
    ax_ACG[0].set_xlim(-.5,.5)
    ax_ACG[0].xaxis.set_ticks([-0.25,0.0,0.25])
    ax_ACG[0].set_ylabel('spike count')
    hist,edges = np.histogram(np.concatenate((isis,isis*-1)),bins=np.linspace(-0.02,0.02,160))
    ax_ACG[1].plot(edges[:-1],hist,drawstyle='steps')
    ax_ACG[1].set_xlim(-.02,.02)
    ax_ACG[1].axvline(0.0015,ls='--',color='#ff8080');ax_ACG[1].axvline(-0.0015,ls='--',color='#ff8080')
    ax_ACG[1].xaxis.set_ticks([-0.02,-0.01,0.0,0.01,0.02])
    for axis in ax_ACG:
        axis.set_xlabel('time (sec)')
    
    
    #text info plot
    cleanAxes(ax_text,total=True)
    ax_text.text(0, 1, 'cluster: '+clusterID, fontsize=12,weight='bold')
    #"%.2f" % signal
    #ax_text.text(10, 30, 'amp.: '+"%.2f" % 1.0+'uV', fontsize=10)
    #ax_text.text(10, 60, 'SNR: '+"%.2f" % float(df[df.clusterID==int(clusterID)].sn_max), fontsize=10)
    #ax_text.text(50, 0, 'isolation distance: '+"%.2f" % float(df[df.clusterID==int(clusterID)].isolation_distance), fontsize=10)
    #ax_text.text(50, 30, 'purity [mahalanobis]: '+"%.2f" % float(1-float(df[df.clusterID==int(clusterID)].mahalanobis_contamination)), fontsize=10)
    #ax_text.text(50, 60, 'ISI violation rate: '+"%.2f" % float(1-float(df[df.clusterID==int(clusterID)].isi_purity)), fontsize=10)
    ax_text.set_ylim(100,0)
    ax_text.set_xlim(0,100)
    
    return plt.gcf()
##===============================================================================
##===============================================================================

# %%



































##===============================================================================
##===============================================================================
##===============================================================================
##===============================================================================
##===============================================================================
##OLD PLOTTING FUNCTIONS
##===============================================================================
##===============================================================================
##===============================================================================
##===============================================================================
def plot_quality(quality,isiV,labels):
    (good,mua,unsorted) = labels
    f,ax = plt.subplots(3,1,figsize=(4,6));
    ax[0].plot(quality[1][np.where(np.in1d(quality[0],good))[0]],quality[2][np.where(np.in1d(quality[0],good))[0]],'o',label='good',mfc='g',mec='g');#plt.ylim(0,1)
    ax[0].plot(quality[1][np.where(np.in1d(quality[0],mua))[0]],quality[2][np.where(np.in1d(quality[0],mua))[0]],'o',label='mua',mfc='r',mec='r');#plt.ylim(0,1)
    ax[0].plot(quality[1][np.where(np.in1d(quality[0],unsorted))[0]],quality[2][np.where(np.in1d(quality[0],unsorted))[0]],'k.',label='unsorted',alpha=0.3);#plt.ylim(0,1)
    ax[0].set_xlabel('iso distance');ax[0].set_ylabel('contamination from mahal.')
    ax[0].set_ylim(ymin=0);ax[0].set_xlim(xmin=1.)
    ax[0].set_xscale('log');
    # if np.max(quality[1]) > 20:
    #     ax[0].set_xlim(1,20)
    legend = ax[0].legend(loc='upper right', shadow=False, fontsize=10,numpoints=1)

    ax[1].plot(quality[1][np.where(np.in1d(quality[0],good))[0]],isiV[1][np.where(np.in1d(isiV[0],good))[0]],'o',label='good',mfc='g',mec='g');plt.ylim(0,1)
    ax[1].plot(quality[1][np.where(np.in1d(quality[0],mua))[0]],isiV[1][np.where(np.in1d(isiV[0],mua))[0]],'o',label='mua',mfc='r',mec='r');plt.ylim(0,1)
    ax[1].plot(quality[1][np.where(np.in1d(quality[0],unsorted))[0]],isiV[1][np.where(np.in1d(isiV[0],unsorted))[0]],'k.',label='unsorted',alpha=0.3);plt.ylim(0,1)
    ax[1].set_xlabel('iso distance');ax[1].set_ylabel('isi contamination')
    ax[1].set_ylim(ymin=0);ax[1].set_xlim(xmin=1.)
    ax[1].set_xscale('log');
    # if np.max(quality[1]) > 20:
    #     ax[1].set_xlim(1,20)
    legend = ax[1].legend(loc='upper right', shadow=False, fontsize=10,numpoints=1)

    ax[2].plot(quality[2][np.where(np.in1d(quality[0],good))[0]],isiV[1][np.where(np.in1d(isiV[0],good))[0]],'o',label='good',mfc='g',mec='g');plt.ylim(0,1)
    ax[2].plot(quality[2][np.where(np.in1d(quality[0],mua))[0]],isiV[1][np.where(np.in1d(isiV[0],mua))[0]],'o',label='mua',mfc='r',mec='r');plt.ylim(0,1)
    ax[2].plot(quality[2][np.where(np.in1d(quality[0],unsorted))[0]],isiV[1][np.where(np.in1d(isiV[0],unsorted))[0]],'k.',label='unsorted',alpha=0.3);plt.ylim(0,1)
    ax[2].set_ylabel('isi contamination');ax[2].set_xlabel('contamination from mahal.')
    ax[2].set_ylim(ymin=0.01);ax[2].set_xlim(xmin=0.01)
    ax[2].set_xscale('log');ax[2].set_yscale('log')
    #legend = ax[2].legend(loc='upper left', shadow=False, fontsize=10,numpoints=1)
    
    plt.tight_layout()

def neuron_fig2(clusterID,df,sortpath,filename='experiment1_102-0_0.dat',time_limits=None,timeplot_binsize=60.,neighbor_colors=["#67572e","#50a874","#ff4d4d"]):
    channels = get_channel_count(sortpath)
    site_positions=option234_positions[np.linspace(0,channels-1,channels).astype(int)]
    data = load_phy_template(sortpath,site_positions)
    
    cluster_ID = int(clusterID)
    cluster_IDs = np.load(os.path.join(sortpath,'spike_clusters.npy'))
    pc_data = np.load(os.path.join(sortpath,'pc_features.npy'))
    pc_ind_data = np.load(os.path.join(sortpath,'pc_feature_ind.npy'))
    params_path = os.path.join(sortpath,'params.py')
    params = read_kilosort_params(params_path)
    spike_times_data = np.load(os.path.join(sortpath,'spike_times.npy'))/ float(params['sample_rate'])
    spike_templates = np.load(os.path.join(sortpath,'spike_templates.npy'))
    
    datapath=os.path.join(sortpath,'experiment1_102-0_0.dat')
    df_line = df[df.clusterID==int(clusterID)]
    
    fig = plt.figure(figsize=(11,8.5))
    ax_text = placeAxesOnGrid(fig,xspan=[0.0,0.4],yspan=[0,0.1])
    ax_position = placeAxesOnGrid(fig,xspan=[0,0.1],yspan=[0.12,1.0])
    ax_waveform = placeAxesOnGrid(fig,xspan=[0.2,0.45],yspan=[0.12,0.65])
    ax_time = placeAxesOnGrid(fig,xspan=[0.2,1.0],yspan=[0.82,1.0])
    ax_PCs = placeAxesOnGrid(fig,xspan=[0.5,0.8],yspan=[0,0.35])
    ax_ACG = placeAxesOnGrid(fig,dim=[1,2],xspan=[0.55,1.0],yspan=[0.5,0.7])
    #ax_neighbor_waveform_1 = placeAxesOnGrid(fig,dim=[1,1],xspan=[0.82,1.0],yspan=[0,0.13])
    #ax_neighbor_waveform_2 = placeAxesOnGrid(fig,dim=[1,1],xspan=[0.82,1.0],yspan=[0.13,0.26])
    #ax_neighbor_waveform_3 = placeAxesOnGrid(fig,dim=[1,1],xspan=[0.82,1.0],yspan=[0.26,0.39])
    ax_neighbor_waveforms=placeAxesOnGrid(fig,dim=[1,1],xspan=[0.82,1.0],yspan=[0.0,0.39])#[ax_neighbor_waveform_1,ax_neighbor_waveform_2,ax_neighbor_waveform_3]
    ax_CCGs_1 = placeAxesOnGrid(fig,dim=[1,1],xspan=[0.53,0.68],yspan=[0.36,0.48])
    ax_CCGs_2 = placeAxesOnGrid(fig,dim=[1,1],xspan=[0.7,.85],yspan=[0.36,0.48])
    ax_CCGs_3 = placeAxesOnGrid(fig,dim=[1,1],xspan=[0.86,1.0],yspan=[0.36,0.48])
    ax_CCGs = [ax_CCGs_1,ax_CCGs_2,ax_CCGs_3]
    
    #position plot
    ax_position.imshow(data[clusterID]['waveform_weights'][::4],extent=(site_positions[:,0][::4][0],site_positions[:,0][::4][0]+16,site_positions[:,1][::4][0],site_positions[:,1][::4][-1]),cmap=plt.cm.gray_r,clim=(0,0.5),interpolation='none')
    ax_position.imshow(data[clusterID]['waveform_weights'][1::4],extent=(site_positions[:,0][1::4][0],site_positions[:,0][1::4][0]+16,site_positions[:,1][1::4][0],site_positions[:,1][1::4][-1]),cmap=plt.cm.gray_r,clim=(0,0.5),interpolation='none')
    ax_position.imshow(data[clusterID]['waveform_weights'][2::4],extent=(site_positions[:,0][2::4][0],site_positions[:,0][2::4][0]+16,site_positions[:,1][2::4][0],site_positions[:,1][2::4][-1]),cmap=plt.cm.gray_r,clim=(0,0.5),interpolation='none')
    ax_position.imshow(data[clusterID]['waveform_weights'][3::4],extent=(site_positions[:,0][3::4][0],site_positions[:,0][3::4][0]+16,site_positions[:,1][3::4][0],site_positions[:,1][3::4][-1]),cmap=plt.cm.gray_r,clim=(0,0.5),interpolation='none')
    ax_position.set_aspect(0.1)
    ax_position.set_ylim(3840,0)
    ax_position.set_xlim(70,0)
    cleanAxes(ax_position)
    ax_position.set_title('neuron position')
    
    #time limits
    if time_limits == None:
        time_limits=[0,1e7]
    all_spikes = data[clusterID]['times']
    spike_times = all_spikes[np.where(all_spikes > time_limits[0])[0][0]:np.where(all_spikes < time_limits[1])[0][-1]]
    #print len(all_spikes)
    #print all_spikes[0]
    #print all_spikes[-1]
    ##print np.where(all_spikes > time_limits[0])[0][0]
    #print np.where(all_spikes < time_limits[1])[0][-1]
    these_spikes = np.where(cluster_IDs==cluster_ID)[0][np.where(all_spikes > time_limits[0])[0][0]:np.where(all_spikes < time_limits[1])[0][-1]]
    #spike_times = spike_times_data[these_spikes]

    #for PC and CCG display, find close by clusters calculate 
    #PC plot
    number_of_spikes_to_plot = 2000
    these_templates=spike_templates[np.where(cluster_IDs==cluster_ID)[0]]
    (included_templates,instances) = count_unique(these_templates)
    this_template = included_templates[int(np.where(instances==np.max(instances))[0])]
    ch1 = pc_ind_data[this_template][0]
    ch2 = pc_ind_data[this_template][1]
    ax_PCs.plot(pc_data[these_spikes][:number_of_spikes_to_plot,0,0],
             pc_data[these_spikes][:number_of_spikes_to_plot,0,1],'bo',ms=1.5,markeredgewidth=0)
    
    nearby_trio = [0,0,0];
    nearby_euclids = [10000,10000,10000];
    nearby_times=[]
    for other_cluster in data.keys():
        if other_cluster != clusterID:
            if (np.abs(data[clusterID]['ypos']-data[other_cluster]['ypos']) + np.abs(data[clusterID]['xpos']-data[other_cluster]['xpos'])) < nearby_euclids[-1]:
                rank = np.where((np.abs(data[clusterID]['ypos']-data[other_cluster]['ypos']) + np.abs(data[clusterID]['xpos']-data[other_cluster]['xpos'])) < nearby_euclids)[0][0]
                nearby_euclids[rank] = (np.abs(data[clusterID]['ypos']-data[other_cluster]['ypos']) + np.abs(data[clusterID]['xpos']-data[other_cluster]['xpos']))
                nearby_trio[rank]=other_cluster
    print(nearby_trio)
    for ii,neighbor in enumerate(nearby_trio):
        all_spikes_neighbor = data[neighbor]['times']
        indices_neighbor= np.where(cluster_IDs==int(neighbor))[0][np.where(all_spikes_neighbor > time_limits[0])[0][0]:np.where(all_spikes_neighbor < time_limits[1])[0][-1]]
        neighbor_templates=spike_templates[indices_neighbor]
        neighbor_spike_times = all_spikes_neighbor[np.where(all_spikes_neighbor > time_limits[0])[0][0]:np.where(all_spikes_neighbor < time_limits[1])[0][-1]]
        nearby_times.append(neighbor_spike_times)
        (included_templates,instances) = count_unique(neighbor_templates)
        this_template = included_templates[int(np.where(instances==np.max(instances))[0])]
        ch1_index = np.where(pc_ind_data[this_template] == ch1)[0]
        ch2_index = np.where(pc_ind_data[this_template] == ch2)[0]
        if ch2_index.size != 0 and ch1_index.size != 0:
            ax_PCs.plot(pc_data[indices_neighbor][:number_of_spikes_to_plot,0,np.where(pc_ind_data[this_template] == ch1)[0][0]],
                     pc_data[indices_neighbor][:number_of_spikes_to_plot,0,np.where(pc_ind_data[this_template] == ch2)[0][0]],
                     'o',color=neighbor_colors[ii],ms=1,markeredgewidth=0,alpha=0.8)
            
            all_diffs = []
            for spike_time in spike_times:
                try:
                    neighbor_start = np.where(neighbor_spike_times < spike_time - 0.5)[0][-1]
                except:
                    neighbor_start = 0
                try:
                    neighbor_end = np.where(neighbor_spike_times > spike_time + 0.5)[0][0]
                except:
                    neighbor_end = -1
                neighbor_chunk = neighbor_spike_times[neighbor_start:neighbor_end]
                #print '\r'+str(spike_time)+' '+str(neighbor_start)+' - '+str(neighbor_end),
                all_diffs.extend(neighbor_chunk - spike_time)
            all_diffs=np.array(all_diffs).flatten()
            hist,edges = np.histogram(all_diffs,bins=np.linspace(-0.2,0.2,400))
            ax_CCGs[ii].plot(edges[:-1],hist,drawstyle='steps',color=neighbor_colors[ii])
            ax_CCGs[ii].set_xlim(-.2,.2)
            ax_CCGs[ii].xaxis.set_ticks([-0.075,0.0,0.075])
            ax_CCGs[ii].axvline(0.0,ls='--',color='#ff8080')
        cleanAxes(ax_CCGs[ii],total=True)
    cleanAxes(ax_neighbor_waveforms,total=True)
    cleanAxes(ax_PCs)    
    ax_PCs.set_title('PC features')
    
    #waveform plot
    cleanAxes(ax_waveform,total=True)
    ax_waveform.set_title('waveform')
    channel_offset = 0 # should not be 0 if not starting from tip.
    peak_y_channel = np.where(data[clusterID]['waveform_weights'] == np.max(data[clusterID]['waveform_weights']))[0][0]
    times = np.random.choice(spike_times,100,replace=False)
    random_times = np.random.rand(100) * (np.max(spike_times)-np.min(spike_times)) + np.min(spike_times)
    xoffs=[0,100,50,150,0,100,50,150,0,100,50,150,0,100,50,150,0,100][::-1]
    yoff=1500
    signal = 0;
    for ii,channel in enumerate(np.linspace(peak_y_channel-8,peak_y_channel+10,18)):
        ws = load_waveforms(datapath,channel,times,channels=channels)
        ws_bkd = load_waveforms(datapath,channel,random_times,channels=channels)
        x_range = np.linspace(xoffs[ii],xoffs[ii]+60,60)
        for i in range(np.shape(ws)[0]):
            if np.shape(x_range)[0] == np.shape(ws[i][:]-yoff*(ii/2))[0]:
                ax_waveform.plot(x_range,ws[i][:]-yoff*(ii/2),alpha=0.05,color='#0066ff')
                ax_waveform.plot(x_range,ws_bkd[i][:]-yoff*(ii/2),alpha=0.05,color='#c8c8c8')
        if np.shape(x_range)[0] == np.shape(np.mean(ws,axis=0)-yoff*(ii/2))[0]:
            if np.max(np.abs(np.mean(ws,axis=0)-yoff*(ii/2))) > signal:
                signal = np.max(np.abs(np.mean(ws,axis=0)-yoff*(ii/2)))
                noise = np.mean(ws_bkd) + np.std(ws_bkd)* 4.
            ax_waveform.plot(x_range,np.mean(ws,axis=0)-yoff*(ii/2),color='#0066ff')
        if ii > 3 and ii < 14:
            for nn,axis in enumerate(nearby_trio):
                    if np.shape(x_range)[0] == np.shape(np.mean(ws,axis=0)-yoff*(ii/2))[0]:
                        neighbor_ws = load_waveforms(datapath,channel,np.random.choice(nearby_times[nn],100,replace=False),channels=channels)
                        ax_neighbor_waveforms.plot(x_range,np.mean(ws,axis=0)-yoff*(ii/2),color='#0066ff')
                        ax_neighbor_waveforms.plot(x_range,np.mean(neighbor_ws,axis=0)-yoff*(ii/2),color=neighbor_colors[nn],alpha=0.8)



    #time plot
    hist,edges = np.histogram(spike_times,bins=np.ceil(spike_times[-1] / timeplot_binsize))
    ax_time.plot(edges[1:],hist/float(timeplot_binsize),drawstyle='steps')  
    ax_time.set_xlabel('time (sec)')
    ax_time.set_ylabel('firing rate (Hz)')
    ax_time.set_title('firing rate over time')
    max_spikes_to_plot = 2000
    if len(spike_times) > max_spikes_to_plot:
        times = np.random.choice(spike_times,max_spikes_to_plot,replace=False)
    else:
        times = spike_times
    amps = get_spike_amplitudes(datapath,peak_y_channel,times)
    ax_time_r = ax_time.twinx()
    ax_time_r.plot(times,amps,'o',markersize=2,alpha=0.1) 
    ax_time_r.set_ylabel('ampltiude')
    cleanAxes(ax_time,leftLabels=True,bottomLabels=True)
    cleanAxes(ax_time_r,rightLabels=True,bottomLabels=True)
    
    #for ACG display, calculate ISIs.
    isis = np.diff(spike_times)
    hist,edges = np.histogram(np.concatenate((isis,isis*-1)),bins=np.linspace(-0.5,0.5,100))
    ax_ACG[0].plot(edges[:-1],hist,drawstyle='steps')
    ax_ACG[0].set_xlim(-.5,.5)
    ax_ACG[0].xaxis.set_ticks([-0.25,0.0,0.25])
    ax_ACG[0].set_ylabel('spike count')
    hist,edges = np.histogram(np.concatenate((isis,isis*-1)),bins=np.linspace(-0.02,0.02,160))
    ax_ACG[1].plot(edges[:-1],hist,drawstyle='steps')
    ax_ACG[1].set_xlim(-.02,.02)
    ax_ACG[1].axvline(0.0015,ls='--',color='#ff8080');ax_ACG[1].axvline(-0.0015,ls='--',color='#ff8080')
    ax_ACG[1].xaxis.set_ticks([-0.02,-0.01,0.0,0.01,0.02])
    for axis in ax_ACG:
        axis.set_xlabel('time (sec)')
    
    
    #text info plot
    cleanAxes(ax_text,total=True)
    ax_text.text(0, 1, 'cluster: '+clusterID, fontsize=12,weight='bold')
    "%.2f" % signal
    ax_text.text(10, 30, 'amp.: '+"%.2f" % signal+'uV', fontsize=10)
    ax_text.text(10, 60, 'SNR: '+"%.2f" % (signal/noise), fontsize=10)
    #ax_text.text(50, 0, 'isolation distance: '+"%.2f" % quality[1][np.where(quality[0]==int(clusterID))[0][0]], fontsize=10)
    #ax_text.text(50, 30, 'purity [mahalanobis]: '+"%.2f" % quality[2][np.where(quality[0]==int(clusterID))[0][0]], fontsize=10)
    #ax_text.text(50, 60, 'ISI violation rate: '+"%.2f" % isiV[1][np.where(isiV[0]==int(clusterID))[0][0]]+'%', fontsize=10)
    ax_text.set_ylim(100,0)
    ax_text.set_xlim(0,100)
    
    return plt.gcf()
##===============================================================================
##===============================================================================
##===============================================================================
##===============================================================================
##===============================================================================


import matplotlib.pyplot as plt
import numpy as np
import scipy.ndimage as ndimage
from dlab.generalephys import placeAxesOnGrid, smoothRF
from matplotlib.patches import Rectangle, Circle
from matplotlib.collections import PatchCollection
import scipy.optimize as opt

#smooth a 2D image, meant to be space-space of a receptive field
#size = number of pixels to smooth over
def smoothRF(img,size=3):
    smooth = ndimage.gaussian_filter(img,(size,size))
    return smooth

from scipy.signal import boxcar,convolve
def smooth_boxcar(data,boxcar_size):
    smoothed = convolve(data,boxcar(boxcar_size))/boxcar_size
    smoothed = smoothed[boxcar_size/2:len(data)+(boxcar_size/2)]
    return smoothed

#show the space-space plots of an already computed STRF for a range of taus.
def plotsta(sta,taus=(np.linspace(-10,280,30).astype(int)),colorrange=(-0.15,0.15),title='',taulabels=False,nrows=3,cmap=plt.cm.seismic,smooth=None,window = [[0,64],[0,64]]):
    ncols = np.ceil(len(taus) / nrows ).astype(int)#+1
    fig,ax = plt.subplots(nrows,ncols,figsize=(10,6))
    titleset=False
    m=np.mean(sta[str(taus[3])])
    for i,tau in enumerate(taus):
        axis = ax[int(np.floor(i/ncols))][i%ncols]
        if smooth is None:
            img = sta[str(tau)].T 
        else:
            img = smoothRF(sta[str(tau)].T,smooth)

        axis.imshow(img,cmap=cmap,vmin=colorrange[0],vmax=colorrange[1],interpolation='none')
        axis.set_frame_on(False);
        axis.set_xticklabels('',visible=False);
        axis.set_xticks([]);
        axis.set_yticklabels('',visible=False);
        axis.set_yticks([])
        axis.set_aspect(1.0)
        axis.set_xlim(window[0][0],window[0][1])
        axis.set_ylim(window[1][0],window[1][1])
        if taulabels:
            axis.set_title('tau = '+str(tau),fontsize=8)
        if titleset is not True:
            axis.set_title(title,fontsize=12)
            titleset=True
        else:
            if tau == 0:
                axis.set_title('tau = '+str(tau),fontsize=12)
    plt.tight_layout()
    fig.show()


#function for fitting with a 2-dimensional gaussian.
def twoD_Gaussian(p, amplitude, xo, yo, sigma_x, sigma_y, theta, offset):
    x=p[0];y=p[1]
    xo = float(xo)
    yo = float(yo)    
    a = (np.cos(theta)**2)/(2*sigma_x**2) + (np.sin(theta)**2)/(2*sigma_y**2)
    b = -(np.sin(2*theta))/(4*sigma_x**2) + (np.sin(2*theta))/(4*sigma_y**2)
    c = (np.sin(theta)**2)/(2*sigma_x**2) + (np.cos(theta)**2)/(2*sigma_y**2)
    g = offset + amplitude*np.exp( - (a*((x-xo)**2) + 2*b*(x-xo)*(y-yo) 
                            + c*((y-yo)**2)))
    return g.ravel()
def fit_rf_2Dgauss(data,center_guess,width_guess=2,height_guess=2):
    dataToFit = data.ravel()
    x=np.linspace(0,np.shape(data)[0]-1,np.shape(data)[0])
    y=np.linspace(0,np.shape(data)[1]-1,np.shape(data)[1])
    x, y = np.meshgrid(x, y)
    popt,pcov = opt.curve_fit(twoD_Gaussian,(x,y),dataToFit,p0=(data[center_guess[1]][center_guess[0]], center_guess[1], center_guess[0], width_guess, height_guess, 0, 0))
    reshaped_to_space=(x,y,twoD_Gaussian((x,y),*popt).reshape(np.shape(data)[1],np.shape(data)[0]))
    return popt,pcov,reshaped_to_space

def fit_rf_2Dgauss_centerFixed(data,center_guess,width_guess=2,height_guess=2):
    dataToFit = data.ravel()
    x=np.linspace(0,np.shape(data)[0]-1,np.shape(data)[0])
    y=np.linspace(0,np.shape(data)[1]-1,np.shape(data)[1])
    x, y = np.meshgrid(x, y)
    
    def twoD_Gaussian_fixed(p, amplitude,sigma_x, sigma_y, theta, offset):
        x=p[0];y=p[1]
        a = (np.cos(theta)**2)/(2*sigma_x**2) + (np.sin(theta)**2)/(2*sigma_y**2)
        b = -(np.sin(2*theta))/(4*sigma_x**2) + (np.sin(2*theta))/(4*sigma_y**2)
        c = (np.sin(theta)**2)/(2*sigma_x**2) + (np.cos(theta)**2)/(2*sigma_y**2)
        g = offset + amplitude*np.exp( - (a*((x-center_guess[1])**2) + 2*b*(x-center_guess[1])*(y-center_guess[0]) 
                                + c*((y-center_guess[0])**2)))
        return g.ravel()
    popt,pcov = opt.curve_fit(twoD_Gaussian_fixed,(x,y),dataToFit,p0=(data[int(center_guess[0])][int(center_guess[1])], width_guess, height_guess, 0, 0))
    reshaped_to_space=(x,y,twoD_Gaussian_fixed((x,y),*popt).reshape(np.shape(data)[1],np.shape(data)[0]))
    return popt,pcov,reshaped_to_space

def impulse(sta,center,taus = np.arange(-10,290,10).astype(int)):
	impulse = [sta[str(tau)][int(center[0])][int(center[1])] for tau in taus]
	return (taus,impulse)


#try to fit an already computed STRF by finding the maximum in space-time, fitting with a 2D gaussian in space-space, and pulling out the temporal kernel at the maximum space-space pixel.
#is very, very finicky right now, requires lots of manual tweaking. 
def fitRF(RF,threshold=None,fit_type='gaussian_2D',verbose=False,rfsizeguess=1.2,flipSpace=False,backup_center=None,zoom_int=10,zoom_order=5,centerfixed=False):
#takes a dictinary containing:
#   
# returns a dictionary containing:
#   the averaged spatial RF, 
#   the centroid of the fit [max fit]
#   a 2D gaussian fit of that spatial RF
#   the impulse response at the center of the fit
#   TODO: a fit of that impulse response with: ?? currently not defined.
    if False:#if np.isnan(RF[RF.keys()[0]][0][0]):#check to make sure there is any data in the STRF to try to fit. if not, return the correct data structure filled with None
        fit={};fit['avg_space_fit']=None;fit['params'] = None;fit['cov']=None ;fit['amplitude']=None ;fit['x']=None ;fit['y']=None ;fit['s_x']=None ;fit['s_y']=None ;fit['theta']=None ;fit['offset']=None;fit['center']=None;fit['peakTau']=None;fit['impulse']=None;fit['roughquality']=None
        return fit
    else:

        
        if type(RF)==dict:
            if 'fit' in RF.keys():
                trash = RF.pop('fit') # get rid of the dictionary entry 'fit'; we only want sta tau frames in this dictionary
                taus = [int(i) for i in RF.keys()]
                taus.sort()
                fit={}
            #========================================================================
            #find the taus to average over for the spatial RF
            #first, define the threshold; above this means there is a non-noise pixel somwhere in the space-space
            if threshold == None:
                #set to z sd above mean
                blank = (RF['-10']+RF['0']+RF['10'])/3. # define a blank, noise-only image
                threshold = np.mean(blank)+np.std(blank)*3.
                if verbose:
                    print('threshold: '+str(threshold))
                
            #find the average space-space over only the range of non-noise good 
            avgRF = np.zeros(np.shape(RF[str(int(taus[0]))]))#initialize the average to blank.
            goodTaus = [40,50,60,70,80,90,100]#
            for tau in goodTaus:
                avgRF += RF[str(int(tau))]
            avgRF = avgRF / float(len(goodTaus))
            fit['avg_space']=avgRF

            maximum_deviation = 0;best_center = (0,0)
            for i in np.linspace(24,63,40):
                for j in np.linspace(10,49,40):
                    imp_temp = impulse(RF,(i,j))
                    if np.max(np.abs(imp_temp[1])) > maximum_deviation:
                        best_center = (int(i),int(j))
                        maximum_deviation = np.max(np.abs(imp_temp[1]))
            center = best_center
            imp_temp = impulse(RF,center)
            if verbose:
                print('peak frame tau: '+str(int(imp_temp[0][np.where(np.array(np.abs(imp_temp[1]))==np.max(np.abs(imp_temp[1])))[0][0]])))
                print('peak center   : '+str(center))
                print('peak value    : '+str(RF[str(int(imp_temp[0][np.where(np.array(np.abs(imp_temp[1]))==np.max(np.abs(imp_temp[1])))[0][0]]))][center[0],center[1]]))
            peak_frame = RF[str(int(imp_temp[0][np.where(np.array(np.abs(imp_temp[1]))==np.max(np.abs(imp_temp[1])))[0][0]]))]
            peak = peak_frame[int(center[0]),int(center[1])]
            #center = (np.where(np.abs(smoothRF(peak_frame,1)) == np.max(np.abs(smoothRF(peak_frame,1))))[0][0],np.where(np.abs(smoothRF(peak_frame,1)) == np.max(np.abs(smoothRF(peak_frame,1))))[1][0])
            
            #========================================================================   
        if type(RF)==np.ndarray:
            fit={}
            fit['avg_space']=RF
            peak_frame = RF
            center = backup_center
            peak = np.max(peak_frame)
        #====fit==================================================================

        if verbose:
            print('peak amp: '+str(peak)+'  threshold: '+str(threshold))
        if np.abs(peak) > threshold * 1.0:
            peak_frame = smoothRF(ndimage.zoom(peak_frame,zoom_int,order=zoom_order),0)
            fit['roughquality']='good'
        else:
            center = backup_center
            imp_temp = impulse(RF,center)
            peak_frame = RF[str(int(100))]
            peak = peak_frame[center[0],center[1]]
            peak_frame = smoothRF(zoom(peak_frame,zoom_int,order=zoom_order),0)
            print('could not find a peak in the RF, using center: '+str(center))
            print('peak amplitude: '+str(peak)+', threshold: '+str(threshold))
            fit['roughquality']='bad'
        fit['center']=center
        fit['center_guess']=center
        fit['fit_image']=peak_frame
        if verbose:
            print('center guess: '+str(center))
        
        #initialize some empty parameters
        fitsuccess=False;retry_fit = False
        best_fit = 10000000#initialize impossibly high
        fit['avg_space_fit']=None;fit['params']=None
        best_fit_output = ((None,None,None,None,None,None,None),600,None)
        try:
            if centerfixed:
                popt,pcov,space_fit = fit_rf_2Dgauss_centerFixed(peak_frame,(center[0]*zoom_int,center[1]*zoom_int),width_guess=rfsizeguess*zoom_int,height_guess=rfsizeguess*zoom_int)
            else:
                popt,pcov,space_fit = fit_rf_2Dgauss(peak_frame,(center[0]*zoom_int,center[1]*zoom_int),width_guess=rfsizeguess*zoom_int,height_guess=rfsizeguess*zoom_int)
        except:
            popt,pcov,space_fit=((None,0,0,0,0,0,0),600,np.zeros((64,64)));print('2D fit failed')
        
        fit['avg_space_fit']=np.array(space_fit)/float(zoom_int)
        fit['params'] = popt    
        fit['cov']=pcov
        fit['amplitude']=popt[0]
        if centerfixed:
            fit['x']=center[1]
            fit['y']=center[0]
            fit['s_x']=popt[1] / float(zoom_int)
            fit['s_y']=popt[2] / float(zoom_int)
            fit['theta']=popt[3]
            fit['offset']=popt[4] / float(zoom_int)
        else:
            fit['x']=popt[1] / float(zoom_int)
            fit['y']=popt[2] / float(zoom_int)
            fit['s_x']=popt[3] / float(zoom_int)
            fit['s_y']=popt[4] / float(zoom_int)
            fit['theta']=popt[5]
            fit['offset']=popt[6] / float(zoom_int)
        #======================================================================== 
    
    #        
        #============get impulse======================================================================== 
        if verbose:
            print('center: '+str(center[0])+' '+str(center[1]))
            
        if fit['avg_space_fit'] is not None:
            center_h = (np.ceil(fit['y']),np.ceil(fit['x']))
            center_r = (np.round(fit['y']),np.round(fit['x']))
            center_l = (np.floor(fit['y']),np.floor(fit['x']))
        try:
            impuls_h = impulse(RF,center_h,taus)[1]
            impuls_r = impulse(RF,center_r,taus)[1]
            impuls_l = impulse(RF,center_l,taus)[1]
            if np.max(np.abs(impuls_h)) > np.max(np.abs(impuls_r)):
                if np.max(np.abs(impuls_h)) > np.max(np.abs(impuls_l)):
                    impuls = impuls_h
                    center = center_h
                else:
                    impuls = impuls_l
                    center= center_l
            else:
                if np.max(np.abs(impuls_r)) > np.max(np.abs(impuls_l)):
                    impuls= impuls_r
                    center= center_r
                else:
                    impuls= impuls_l
                    center= center_l
        except:
            impuls = np.array([0,0,0,0,0,0]) #impuls = np.zeros(len(taus))
            taus = impuls
        if fit_type == 'gaussian_2D':
            #get impulse at the 'center'
            if verbose:
                print('center from fit: '+str(center[0])+' '+str(center[1]))
            #impuls = [RF[str(tau)][center[0]][center[1]] for tau in taus]
            fit['impulse']=(np.array(taus),np.array(impuls))
            peakTau = taus[np.abs(np.array(impuls)).argmax()]
            peakTau = 80
            fit['peakTau']=peakTau
        
        fit['center_usedforfit'] = fit['center']
        fit['center_usedforimp'] = center
        fit['impulse']=(np.array(taus),np.array(impuls))
        #======================================================================== 
        
        return fit 


#convenience plotting method for showing spatial and temporal filters pulled from fitting an already computed STRF
def show_sta_fit(fit,colorrange=(0.35,0.65),cmap=plt.cm.seismic,title='',contour_levels=3):
    
    if fit is not None:
        fig = plt.figure(figsize=(6,2.75))
        ax_full_space = placeAxesOnGrid(fig,dim=(1,1),xspan=(0,0.32),yspan=(0,0.5))
        ax_zoom_space = placeAxesOnGrid(fig,dim=(1,1),xspan=(0.33,0.63),yspan=(0,0.5))
        ax_zoom_space_filtered = placeAxesOnGrid(fig,dim=(1,1),xspan=(0.56,0.86),yspan=(0,0.5))
        ax_impulse = placeAxesOnGrid(fig,dim=(1,1),xspan=(0.1,0.7),yspan=(0.54,1.0))
        
        ax_impulse.plot(fit['impulse'][0],np.zeros(len(fit['impulse'][0]))+0.5,'r-')
        ax_impulse.plot(fit['impulse'][0],fit['impulse'][1]);
        ax_impulse.set_ylim(colorrange[0],colorrange[1]) 
        ax_impulse.set_xlabel('time (msec)') 
        ax_impulse.set_ylabel('spike probability')
        #ax_impulse.text(1,1,str(np.mean(fit['cov'])))
        
        ax_full_space.plot([fit['y']],[fit['x']],'k+',markersize=6.0)#([fit['center'][0]],[fit['center'][1]],'k+',markersize=6.0)
        ax_full_space.imshow(fit['avg_space'].T,
           cmap=cmap,
           vmin=colorrange[0],vmax=colorrange[1],
            interpolation='none');
        plt.setp(ax_full_space.get_xticklabels(),visible=False);ax_zoom_space_filtered.xaxis.set_major_locator(plt.NullLocator())
        plt.setp(ax_full_space.get_yticklabels(),visible=False);ax_zoom_space_filtered.yaxis.set_major_locator(plt.NullLocator())
        ax_full_space.set_title(title)
        ax_full_space.set_ylim(20,40)
        ax_full_space.set_xlim(12,52)
        ax_full_space.axvline(x=32,linewidth=1,color='r',ls='--')
        
        zoom_size = 5 #number of pixels in each direction from peak center to keep in zooms
        ax_zoom_space.imshow(fit['avg_space'],
                   cmap=cmap,
                   vmin=colorrange[0],vmax=colorrange[1],
                    interpolation='none');
        #ax_zoom_space.plot([fit['x']],[fit['y']],'k+',markersize=6.0)
        ax_zoom_space.set_xlim(fit['center'][1]-zoom_size,fit['center'][1]+zoom_size)
        ax_zoom_space.set_ylim(fit['center'][0]-zoom_size,fit['center'][0]+zoom_size) 
        plt.setp(ax_zoom_space.get_xticklabels(),visible=False);ax_zoom_space_filtered.xaxis.set_major_locator(plt.NullLocator())
        plt.setp(ax_zoom_space.get_yticklabels(),visible=False);ax_zoom_space_filtered.yaxis.set_major_locator(plt.NullLocator())
        

        ax_zoom_space_filtered.imshow(smoothRF(fit['avg_space'],1),
                   cmap=cmap,
                   vmin=colorrange[0]/3.,vmax=colorrange[1]/3.,
                    interpolation='none');
        
        #ax_zoom_space_filtered.plot([fit['x']],[fit['y']],'k+',markersize=6.0)
        ax_zoom_space_filtered.set_xlim(fit['center'][1]-zoom_size,fit['center'][1]+zoom_size)
        ax_zoom_space_filtered.set_ylim(fit['center'][0]-zoom_size,fit['center'][0]+zoom_size)
        plt.setp(ax_zoom_space_filtered.get_xticklabels(),visible=False);ax_zoom_space_filtered.xaxis.set_major_locator(plt.NullLocator())
        plt.setp(ax_zoom_space_filtered.get_yticklabels(),visible=False);ax_zoom_space_filtered.yaxis.set_major_locator(plt.NullLocator())
        
    
        

        
        if 'avg_space_fit' in fit.keys():
            if fit['avg_space_fit'] is not None:
                if np.max(np.abs(fit['impulse'][1])) > np.std(fit['impulse'][1])*2.:
                    ax_full_space.add_patch(Rectangle((fit['y']-zoom_size,fit['x']-zoom_size),zoom_size*2,zoom_size*2,fill=None,ls='dotted'))
                    ax_zoom_space.contour(fit['avg_space_fit'][0],
                                fit['avg_space_fit'][1],
                                fit['avg_space_fit'][2],
                                contour_levels)
                    ax_zoom_space_filtered.contour(fit['avg_space_fit'][0],
                                fit['avg_space_fit'][1],
                                fit['avg_space_fit'][2],
                                contour_levels)
                    ax_full_space.contour(fit['avg_space_fit'][1],
                                fit['avg_space_fit'][0],
                                fit['avg_space_fit'][2],
                                contour_levels)
                    ax_full_space.set_aspect('equal')   

		#                   vmin=-0.08,vmax=0.08);
        #plt.tight_layout()
        return fig
    

#compute a spike-triggered average on three dimensional data. this is typically a movie of the stimulus
#TODO: should be modified to take data of any shape (for example, an LFP trace) to average.
def sta(spiketimes,data,datatimes,taus=(np.linspace(-10,280,30)),exclusion=None,samplingRateInkHz=25):
    output = {}
    for tau in taus:
        avg = np.zeros(np.shape(data[:,:,0]))
        count = 0
        for spiketime in spiketimes:
            if spiketime > datatimes[0] and spiketime < datatimes[-1]-0.6:
                if exclusion is not None: #check to see if there is a period we are supposed to be ignoring, because of eye closing or otherwise
                    if spiketime > datatimes[0] and spiketime < datatimes[-1]-0.6:
                         index = (np.where(datatimes > (spiketime - tau*samplingRateInkHz))[0][0]-1) % np.shape(data)[2]
                         avg += data[:,:,index]
                else:
                	index = (np.where(datatimes > (spiketime - tau*samplingRateInkHz))[0][0]-1) % np.shape(data)[2]
                	avg += data[:,:,index]
                count+=1
        output[str(int(tau))]=avg/count
    return output

#should compute a spike-triggered average on n-dimensional data.
#this is typically a movie of the stimulus of shape [frames,x,y], but could be any shape, such as a 1-d noise stimulus or continuous trace.
#as of now, limited to computation of the average along the first axis of 'data'. 
def sta2(spiketimes,data,datatimes,taus=(np.linspace(-10,280,30)),exclusion=None,samplingRateInkHz=30,time_domain=False):
    output = {}
    if time_domain:
        taus = taus / 1000.
    else:
        taus = (taus * samplingRateInkHz)/1000.
		
    for tau in taus:
        avg = np.zeros(np.shape(data[0]))
        count = 0
        for spiketime in spiketimes:
            if spiketime > datatimes[0] and spiketime < datatimes[-1]-0.5:
                if exclusion is not None: #check to see if there is a period we are supposed to be ignoring, because of eye closing or otherwise
                    if spiketime > datatimes[0] and spiketime < datatimes[-1]-0.5:
                         index = (np.where(datatimes > (spiketime - tau))[0][0]-1) % np.shape(data)[0]
                         avg += data[index]
                else:
                	index = (np.where(datatimes > (spiketime - tau))[0][0]-1) % np.shape(data)[0]
                	avg += data[index]
                count+=1
        output[str(tau)]=avg/count
    return output

def sta_with_subfields(spiketimes,data,datatimes,taus=(np.linspace(-10,280,30)),exclusion=None,samplingRateInkHz=25):
    output = {}
    output_ON = {}
    output_OFF = {}
    for tau in taus:
        avg = np.zeros(np.shape(data[:,:,0]))
        avg_ON = np.zeros(np.shape(data[:,:,0]))
        avg_OFF = np.zeros(np.shape(data[:,:,0]))
        count = 0
        for spiketime in spiketimes:
            if spiketime > datatimes[0] and spiketime < datatimes[-1]-0.6:
                if exclusion is not None: #check to see if there is a period we are supposed to be ignoring, because of eye closing or otherwise
                    if spiketime > datatimes[0] and spiketime < datatimes[-1]-0.6:
                         index = (np.where(datatimes > (spiketime - tau*samplingRateInkHz))[0][0]-1) % np.shape(data)[2]
                         avg += data[:,:,index]
                         avg_ON += (data[:,:,index] + 1 ) / 2
                         avg_OFF += (data[:,:,index]- 1 ) / 2
                else:
                	index = (np.where(datatimes > (spiketime - tau*samplingRateInkHz))[0][0]-1) % np.shape(data)[2]
                	avg += data[:,:,index];avg_ON += (data[:,:,index] + 1 ) / 2;avg_OFF += (data[:,:,index] - 1 ) / 2
                count+=1
        output[str(int(tau))]=avg/count
        output_ON[str(int(tau))]=avg_ON/count# - 0.5
        output_OFF[str(int(tau))]=avg_OFF/count# + 0.5
    return output,output_ON,output_OFF

#from https://github.com/JesseLivezey/gabor_fit/blob/master/gabor_fit/fit.py
def generate_gabor(p, pixels_x,pixels_y, theta, stdx, stdy, lamb, phase):
    """
    Generate a gabor filter based on given parameters.
    Parameters
    ----------
    pixels : tuple of ints
        Height and width of patch.
    p : (x,y)
        x : float
            x Location of center of gabor in pixels.
        y : float
            y Location of center of gabor in pixels.
    theta : float
        Rotation of gabor in plane in degrees.
    stdx : float
        Width of gaussian window along rot(x) in pixels.
    stdy : float
        Width of gaussian window along rot(y) in pixels.
    lamb : float
        Wavelength of sine funtion in pixels along rot(x).
    phase : float
        Phase of sine function in degrees.
    Returns
    -------
    gabor : ndarray
        2d array of pixel values.
    """
    #print (pixels_x,pixels_y)
    x=p[0];y=p[1]
    x_coords = np.arange(0, int(pixels_x))#pixels[0])
    y_coords = np.arange(0, int(pixels_y))#pixels[1])
    xx, yy = np.meshgrid(x_coords, y_coords)
    unit2rad = 2. * np.pi 
    deg2rad = 2. * np.pi / 360.
    xp = (xx - x) * np.cos(deg2rad * theta) - (yy - y) * np.sin(deg2rad * theta)
    yp = (xx - x) * np.sin(deg2rad * theta) - (yy - y) * np.cos(deg2rad * theta)
    gabor = (np.exp(-xp**2 / (2. * stdx**2) - yp**2 / (2. * stdy**2)) *
             np.sin(unit2rad * (xp / lamb) - deg2rad * phase))
    norm = np.sqrt((gabor**2).sum())
    g = gabor/norm
    return g.ravel()


def fitgabor_2(data,pixels=(10,10),x=32.,y=32.,theta=0.,stdx=3.,stdy=3.,lamb=1.5,phase=0.):

    popt,pcov = opt.curve_fit(generate_gabor,(x,y),data.ravel(),p0=(int(pixels[0]),int(pixels[1]),theta,stdx,stdy,lamb,phase))
    reshaped_to_space=(x,y,generate_gabor((x,y),*popt).reshape(np.shape(data)[1],np.shape(data)[0]))
    return popt,reshaped_to_space,pcov

def check_rfs_in_df(df_rf,sds=4):
    # this function will check if the RFs computed and stored in this dataframe have signal in them, adding three rows to the input df:
    #      rf_computed: if any RF STA was calculated at all
    #      good: if there was a non-noise RF
    #      rf_color: the color of the noise used to generate the "good" RF, if there was a good one
    #df_rf: a pandas DataFrame with some receptive fields computed from uv and green noise to check. 
    #sds: number of SDs a pixels has to be above in order to be considered a real RF pixel
    x=sds
    rf_computed = []
    good_rf = []
    rf_color = []
    for ind in df_rf.index:
        if type(df_rf.g_avg_space[ind]) != str:
            # measure some things about the computed rf
            rf_computed.extend([True])
            
            gr = df_rf.g_fit_image[ind]
            im = smoothRF(gr,.5);
            thresh = np.mean(im)+x*np.std(im)
            im[np.abs(smoothRF(gr,.6))<thresh]=0
            gr_pixels = np.nansum(np.abs(im))
            
            uv = df_rf.u_fit_image[ind]
            im = smoothRF(uv,.5);
            thresh = np.mean(im)+x*np.std(im)
            im[np.abs(smoothRF(uv,.6))<thresh]=0
            uv_pixels = np.nansum(np.abs(im))
    
            if gr_pixels>0 or uv_pixels>0:
                good_rf.extend([True])
                if uv_pixels>gr_pixels:
                    im = smoothRF(uv,.5)
                    rf_color.extend(['uv'])
                else:
                    im = smoothRF(gr,.5)
                    rf_color.extend(['green'])
            else: 
                rf_color.extend([np.nan]);good_rf.extend([False])
        else:
            rf_computed.extend([False])
            good_rf.extend([False])
            rf_color.extend([np.nan])
    df_rf['rf_computed']=rf_computed
    df_rf['good_rf']=good_rf
    df_rf['rf_color']=rf_color
    
    return df_rf

def show_impulse(a,center):
	center=(41,32)
	i = impulse(a,center,taus=np.linspace(-10,280,30).astype(int))
	plt.plot(i[0],i[1])
	plt.ylim(108,148)
	plt.gca().axhline(128,ls='--')
	plt.figure()
	plt.imshow(a['80'])
	plt.gca().scatter(center[1],center[0],color='w',alpha=0.3)



import numpy as np
import matplotlib.pyplot as plt
from dlab.generalephys import color50

#compute and optionally plot a peri-stimulus time histogram
#plot is a line plot, with options for error display [bars or shaded]
def psth_line(times,triggers,pre=0.5,timeDomain=True,post=1,binsize=0.05,ymax=75,yoffset=0,output='fig',name='',color='#00cc00',linewidth=0.5,axes=None,labels=True,sparse=False,labelsize=18,axis_labelsize=20,error='shaded',alpha=0.5,**kwargs):
    post = post + 1
    peris=[]#np.zeros(len(triggers),len(times))
    p=[]
    if timeDomain:
        samplingRate = 1.0
    else:
        samplingRate = samplingRate
        
    times = np.array(times).astype(float) / samplingRate + pre
    triggers = np.array(triggers).astype(float) / samplingRate

    numbins = int((post+pre) / binsize) 
    bytrial = np.zeros((len(triggers),numbins))
    for i,t in enumerate(triggers):
        
        if len(np.where(times >= t - pre)[0]) > 0 and len(np.where(times >= t + post)[0]) > 0:
            start = np.where(times >= t - pre)[0][0]
            end = np.where(times >= t + post)[0][0]
            for trial_spike in times[start:end-1]:
                if float(trial_spike-t)/float(binsize) < float(numbins):
                    bytrial[i][int((trial_spike-t)/binsize-1)] +=1   
        else:
        	 pass
             #bytrial[i][:]=0
        #print 'start: ' + str(start)+'   end: ' + str(end)

    variance = np.std(bytrial,axis=0)/binsize/np.sqrt((len(triggers)))
    hist = np.mean(bytrial,axis=0)/binsize
    edges = np.linspace(-pre+binsize,post+binsize,numbins)

    if output == 'fig':
        if error == 'shaded':
            if 'shade_color' in kwargs.keys():
                shade_color=kwargs['shade_color']
            else:
                shade_color=color    
            if axes == None:
                plt.figure()
                axes=plt.gca()
            plt.locator_params(axis='y',nbins=4)
            upper = hist+variance
            lower = hist-variance
            axes.fill_between(edges[2:-1],upper[2:-1]+yoffset,hist[2:-1]+yoffset,alpha=alpha,color=shade_color,facecolor=shade_color)
            axes.fill_between(edges[2:-1],hist[2:-1]+yoffset,lower[2:-1]+yoffset,alpha=alpha,color=shade_color,facecolor=shade_color)
            axes.plot(edges[2:-1],hist[2:-1]+yoffset,color=color,linewidth=linewidth)
            axes.set_xlim(-pre,post-1)
            axes.set_ylim(0,ymax);
            if sparse:
                axes.set_xticklabels([])
                axes.set_yticklabels([])
            else:
                if labels:
                    axes.set_xlabel(r'$time \/ [s]$',fontsize=axis_labelsize)
                    axes.set_ylabel(r'$firing \/ rate \/ [Hz]$',fontsize=axis_labelsize)
                    axes.tick_params(axis='both',labelsize=labelsize)
            axes.spines['top'].set_visible(False);axes.yaxis.set_ticks_position('left')
            axes.spines['right'].set_visible(False);axes.xaxis.set_ticks_position('bottom')   
            axes.set_title(name,y=1)
            return axes 
        else:
            if axes == None:
                plt.figure()
                axes=plt.gca()
            f=axes.errorbar(edges,hist,yerr=variance,color=color)
            axes.set_xlim(-pre,post - 1)
            axes.set_ylim(0,ymax)
            if sparse:
                axes.set_xticklabels([])
                axes.set_yticklabels([])
            else:
                if labels:
                    axes.set_xlabel(r'$time \/ [s]$',fontsize=axis_labelsize)
                    axes.set_ylabel(r'$firing \/ rate \/ [Hz]$',fontsize=axis_labelsize)
                    axes.tick_params(axis='both',labelsize=labelsize)
            axes.spines['top'].set_visible(False);axes.yaxis.set_ticks_position('left')
            axes.spines['right'].set_visible(False);axes.xaxis.set_ticks_position('bottom')   
            axes.set_title(name)
            return axes
    if output == 'hist':
        return (hist[:-1*int(1./binsize)],edges[:-1*int(1./binsize)])    
    if output == 'p':
        return (edges,hist,variance)
	
#find the x and y position of an average waveform, given the probe geometry
#plot is a bar plot
def psth(times,triggers,timeDomain=False,pre=0.5,post=1,binsize=0.05,ymax=75,output='fig',name='',color=[1,1,1],axes=None,labels=True,sparse=False,labelsize=18):
    peris=[]#np.zeros(len(triggers),len(times))
    if timeDomain:
        samplingRate = 1.0
    else:
        samplingRate = samplingRate
        
    times = np.array(times).astype(float) / samplingRate
    triggers = np.array(triggers).astype(float) / samplingRate
    
    for i,t in enumerate(triggers):
        peris.append(np.array(times).astype(float)-float(t))
    peris = np.array(peris)
    peris=peris.flatten()

    numbins = (post+pre) / binsize 
    (hist,edges) = np.histogram(peris,int(numbins),(-pre,post))
    hist /= float(len(triggers)*binsize)


    if output == 'fig':
        if axes == None:
            plt.figure()
            axes=plt.gca()
        f=axes.bar(edges[:-1],hist,width=binsize,color=color)
        axes.set_xlim(-pre,post)
        axes.set_ylim(0,ymax)
        if sparse:
            axes.set_xticklabels([])
            axes.set_yticklabels([])
        else:
            if labels:
                axes.set_xlabel(r'$time \/ [s]$',fontsize=20)
                axes.set_ylabel(r'$firing \/ rate \/ [Hz]$',fontsize=20)
                axes.tick_params(axis='both',labelsize=labelsize)
        axes.spines['top'].set_visible(False);axes.yaxis.set_ticks_position('left')
        axes.spines['right'].set_visible(False);axes.xaxis.set_ticks_position('bottom')
        axes.set_title(name)
        return axes
    if output == 'hist':
        return (hist,edges)    
    if output == 'p':
        return peris

#compute and optionally show peri-stimulus time histograms for a list of arrays of times
def psthlist(timesdict,timeslist,onsets,pre=0.5,post=1,binsize=0.05,ymax=50,output='fig'):
    fig,(ax1,ax2,ax3,ax4)=plt.subplots(1,np.shape(timeslist)[0],sharey=True) 
    axes=(ax1,ax2,ax3,ax4)
    for j,s in enumerate(timeslist):
        peris=[]#np.zeros(len(onsets),len(times))
        times = timesdict[s]
        times = np.array(times).astype(float) / samplingRate
        onsets = np.array(onsets).astype(float) / samplingRate
        
        for i,t in enumerate(onsets):
            peris.append(np.array(times).astype(float)-float(t))
        peris = np.array(peris) 
        peris=peris.flatten()
    
        numbins = (post+pre) / binsize 
        (hist,edges) = np.histogram(peris,int(numbins),(-pre,post))
        
        hist /= float(len(onsets)*binsize)
    
        print(j)
        axis = axes[j]
        print(axis)
        axis.hist(peris,int(numbins),(-pre,post),color=[1,1,1])
        axis.set_xlim(-pre,post)
        axis.set_ylim(0,ymax)
        if j==0:
            axis.set_xlabel(r'$time \/ [s]$',fontsize=14)
            axis.set_ylabel(r'$firing \/ rate \/ [Hz]$',fontsize=14)
        axis.set_title(s)
    
    plt.show()    
    return fig

def raster(times,triggers,pre=0.5,timeDomain=False,post=1,yoffset=0,output='fig',name='',color='#00cc00',linewidth=0.5,axes=None,labels=True,sparse=False,labelsize=18,axis_labelsize=20,error='',alpha=0.5,ms=2,**kwargs):
    #post = post + 1
    if timeDomain:
        samplingRate = 1.0
    else:
        samplingRate = samplingRate
        
    times = np.array(times).astype(float) / samplingRate + pre
    triggers = np.array(triggers).astype(float) / samplingRate
    bytrial = [];
    
    if axes == None and output!='data':
        plt.figure()
        axes=plt.gca()
        
    for i,t in enumerate(triggers):
        if len(np.where(times >= t - pre - post)[0]) > 0 and len(np.where(times >= t + post+ pre)[0]) > 0:
            start = np.where(times >= t - pre)[0][0]
            end = np.where(times >= t + post)[0][0]
            bytrial.append(np.array(times[start:end])-t-pre)
            if output!='data':
            #		print np.ones(len(np.array(times[start:end-1])-t))*i+1
                axes.plot(np.array(times[start:end])-t-pre,
                          np.ones(len(np.array(times[start:end])-t))*i+1,
                          "|",mew=linewidth,ms=ms,color=color)
        else: bytrial.append([])
                
                
    if output!='data':
        axes.set_xlim(-pre,post)
        axes.set_title(name)
        axes.set_ylim(len(triggers),1)
        if sparse:
            cleanAxes(axes,total=True)
        else:
            if labels:
                axes.set_xlabel(r'$time \/ [s]$',fontsize=16)
                axes.set_ylabel(r'$trial \/ number$',fontsize=16)
                axes.tick_params(axis='both',labelsize=labelsize)
                axes.spines['top'].set_visible(False);axes.yaxis.set_ticks_position('left')
                axes.spines['right'].set_visible(False);axes.xaxis.set_ticks_position('bottom')
    if output == 'fig':
        return (plt.gcf(),plt.gca())
    if output=='data':
        return bytrial


def raster_singletrial(nwb_data,trigger,pre=0.5,timeDomain=True,post=.1,insertion_angle=45,yoffset=0,output='fig',name='',color='#00cc00',linewidth=0.5,axes=None,labels=True,sparse=False,labelsize=18,axis_labelsize=20,error='',alpha=0.5,**kwargs):
	post = post + 1
	if timeDomain:
		samplingRate = 1.0
	else:
		samplingRate = samplingRate
        
	t = float(trigger) / samplingRate
	bycell = [];
	
	if axes == None:
		plt.figure()
		axes=plt.gca()
	for ii,probe in enumerate(nwb_data['processing'].keys()):
		if 'UnitTimes' in nwb_data['processing'][probe]:
			for i,cell in enumerate(np.sort(nwb_data['processing'][probe]['UnitTimes'].keys())[:-1]):
				times = nwb_data['processing'][probe]['UnitTimes'][cell]['times']
				if len(np.where(times >= t - pre)[0]) > 0 and len(np.where(times >= t + post)[0]) > 0:
					start = np.where(times >= t - pre)[0][0]
					end = np.where(times >= t + post)[0][0]
					bycell.append(np.array(times[start:end-1])-t)
					axes.plot(np.array(times[start:end-1])-t,
							  np.ones(len(np.array(times[start:end-1])-t))*(nwb_data['processing'][probe]['UnitTimes'][cell]['ypos']*np.sin(np.deg2rad(90-insertion_angle))),
							  '|',
							  linewidth=1,mew=0.5,
							  #color='#d9d9d9')
							  color=color50[ii%50])
	axes.set_xlim(-pre,post-1.)
	axes.set_ylim(1000,0)
	axes.set_title(name)

	
	if sparse:
		cleanAxes(axes,total=True)
	else:
		if labels:
			axes.set_xlabel(r'$time \/ [s]$',fontsize=16)
			axes.set_ylabel(r'$depth \/ um$',fontsize=16)
			axes.tick_params(axis='both',labelsize=labelsize)
			axes.spines['top'].set_visible(False);axes.yaxis.set_ticks_position('left')
			axes.spines['right'].set_visible(False);axes.xaxis.set_ticks_position('bottom')
	if output == 'fig':
		return (plt.gcf(),plt.gca())
	if output=='data':
		return bycell
        
        


#compute the tuning over a given parameter from a PSTH
def psth_tuning(data,unit,param,params,paramtimesdict,window=1.33,binsize=0.02,savepsth=False,path=r'C:\Users\danield\Desktop\data'):
    tun_y = np.zeros(len(params))
    tun_x = np.zeros(len(params))
    i=0
    for p in params:
        print(p)
        if savepsth:
            f=psth(data[unit]['times'],paramtimesdict[param+str(p)],pre=0,post=window,binsize=binsize)
            f.savefig(os.path.join(path,'unit'+unit+param+str(p)+'_psth.eps'),format='eps')     
        
        (hist,edges) = psth(data[unit]['times'],paramtimesdict[param+str(p)],pre=0,post=window,binsize=binsize,output='hist')        
        tun_y[i] = np.mean(hist)
        tun_x[i] = p
        i+=1
        
    plt.plot(tun_x,tun_y,'ko-');
    plt.xscale('log');
    plt.xlim(7,101);
    plt.ylabel(r'$firing \/ rate \/ [Hz]$',fontsize=14);
    plt.xlabel(r'$contrast \/ $[%]',fontsize=14);
    f=plt.gcf()
    return f

#compute the latency to first reseponse from a PSTH
def psth_latency(data,bins,pre=None,binsize=None, sd = 2.5,smooth=False,offset=0):
    if smooth:
        data = scipy.signal.savgol_filter(data,5,3)
    if pre is None:
        pre = bins[0]
    if binsize == None:
        binsize = bins[1]-bins[0]
    startbin = np.where(bins>0)[0][0]
    baseline = np.mean(data[:startbin])
    threshold = baseline + np.std(data[:startbin])*sd +0.2
    crossings = plt.mlab.cross_from_below(data[startbin:],threshold)
    if len(crossings)>0:
        crossing = crossings[0]#the first bin above the threshold
        chunk = np.linspace(data[crossing+startbin-1],data[crossing+startbin],100)
        bin_crossing = plt.mlab.cross_from_below(chunk,threshold)
        latency =(crossing-1)*(1000*binsize)+bin_crossing/100.0 * (1000*binsize) 
    else:
        #print 'response did not exceed threshold: '+str(threshold)+', no latency returned'
        return None
    return latency[0] - offset

#compute the latency to first reseponse from a PSTH
def psth_area(data,bins,pre=None,binsize=None, sd = 3,time=0.2):
    if pre is None:
        pre = bins[0]
    if binsize == None:
        binsize = bins[1]-bins[0]
    startbin = np.where(bins>0)[0][0]
    baseline = np.mean(data[:startbin])
    threshold = baseline + np.std(data[:startbin])*sd +0.2
    crossings = plt.mlab.cross_from_below(data[startbin:],threshold)
    if len(crossings)>0:
        try:
            area = np.trapz(np.abs(data[startbin:startbin+np.ceil(time/binsize)]) - baseline)
            return area
        except: return None 
        print('response did not exceed threshold: '+str(threshold)+', no area returned')
        return None

def psth_arr(spike_data, unit, stim_data, condition, pre=0.5, post=2.5,binsize=0.05,variance=True):
    times = np.array(spike_data[spike_data.unit_id==unit].times.values[0])
    numbins = int((post+pre)/binsize)
    conds = np.unique(stim_data[condition])
    num_conds = len(conds)
    x = np.arange(-pre,post,binsize)
    colors = plt.cm.viridis(np.linspace(0,1,num_conds))
    
    psth_all=[]
    bytrial_all=[]
    var_all = []
    
    for i,cond in enumerate(np.unique(stim_data[condition])):
        triggers = np.array(stim_data['times'][stim_data[condition] == cond])
#         print(triggers.shape)
        bytrial = np.zeros((len(triggers),numbins-1))
        for j, trigger in enumerate(triggers):
            trial = triggers[j]
            start = trial-pre
            end = trial+post
            bins_ = np.arange(start,end,binsize)
            trial_spikes = times[np.logical_and(times>=start, times<=end)]
            hist,edges = np.histogram(trial_spikes,bins=bins_)
            if len(hist)==numbins-1:
                bytrial[j]=hist
            elif len(hist)==numbins:
                bytrial[j]=hist[:-1]
        if variance == True:
            var = np.std(bytrial,axis=0)/binsize/np.sqrt((len(triggers)))
            psth = np.nanmean(bytrial,axis=0)/binsize
            var_all.append(var)
        psth_all.append(psth)
        bytrial_all.append(bytrial)
    bytrial_all = dict(zip(np.unique(stim_data[condition]),bytrial_all))
    psth_all = dict(zip(np.unique(stim_data[condition]),psth_all))
    var_all = dict(zip(np.unique(stim_data[condition]),var_all))

    return(psth_all,bytrial_all,var_all)

def psth_line_overlay_(spike_data, unit, stim_data, condition, title='', 
                       pre=0.5, post=2.5,binsize=0.05,variance=True,axis=None,legend=True):
#     times = np.array(np.array(spike_data.times[spike_data.unit_id==unit])[0])
    times = np.array(spike_data[spike_data.unit_id==unit].times.values[0])
    numbins = int((post+pre)/binsize)
    conds = np.unique(stim_data[condition])
    num_conds = len(conds)
    x = np.arange(-pre,post,binsize)
    colors = plt.cm.viridis(np.linspace(0,1,num_conds))

    psth_all=[]

    if axis == None:
        fig,ax = plt.subplots()
    else:
        ax = axis; fig = plt.gcf()

    for i,cond in enumerate(np.unique(stim_data[condition])):
        triggers = np.array(stim_data['times'][stim_data[condition] == cond])
        bytrial = np.zeros((len(triggers),numbins-1))
        for j, trigger in enumerate(triggers):
            trial = triggers[j]
            start = trial-pre
            end = trial+post
            bins_ = np.arange(start,end,binsize)
            trial_spikes = times[np.logical_and(times>=start, times<=end)]
            hist,edges = np.histogram(trial_spikes,bins=bins_)
            if len(hist)==numbins-1:
                bytrial[j]=hist
            elif len(hist)==numbins:
                bytrial[j]=hist[:-1]
        psth = np.mean(bytrial,axis=0)/binsize
        if isinstance(conds[i],float)==True:
            ax.plot(x[:-1],psth, color=colors[i], label=str(round(conds[i],2)))
#         if isinstance(conds[i],tuple)==True:
#             ax.plot(x[:-1],psth, color=colors[i], label=str(round(conds[i],2)))
        else:
            ax.plot(x[:-1],psth, color=colors[i], label=str(conds[i]))
        if variance == True:
            var = np.std(bytrial,axis=0)/binsize/np.sqrt((len(triggers)))
            upper = psth+var
            lower = psth-var
            ax.fill_between(x[:-1],upper,psth,alpha=0.1,color=colors[i])
            ax.fill_between(x[:-1],lower,psth,alpha=0.1,color=colors[i])
    ax.axvline(0,linestyle='dashed')
    if legend==True:
        plt.legend(loc=(1.05,0.48))
    plt.title(title)
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
#     return(ax)



#general imports
import os,sys,glob
import numpy as np
import pandas as pd
import h5py
import matplotlib.pyplot as plt
from matplotlib.patches import Rectangle
import seaborn as sns
import csv 
import time, datetime
# ! pip install git+https://github.com/open-ephys/open-ephys-python-tools
from open_ephys.analysis import Session
from open_ephys.control import NetworkControl
import numpy as np
import matplotlib.pyplot as plt


def electrical_timestamps(path,rec_done,ec):
    
    #Pulls Wanted Data from directory location, turns it into session
    directory = path
    session = Session(directory)
    recordnode = session.recordnodes[0]
    data = session.recordnodes[0].recordings[0].continuous
    sampling_rate = 30000.
    analog_channels = data[0].samples
    times = data[0].timestamps/sampling_rate
    metadata = data[0].metadata  
    ##
    num_rec_done = list(range(rec_done))
    num_rec_done_a = np.array(num_rec_done)
    rec_num = rec_done
    # WRS
    w_high_samples = np.where(analog_channels[:,ec]>100)[0]
    w_low_samples = np.where(analog_channels[:,ec]<-100)[0]
    
    wrs = []
    wfs = []
    for transition in np.where(np.diff(w_high_samples)>10)[0]:
        wrs.extend([w_high_samples[transition]]) 
    for transition in np.where(np.diff(w_low_samples)>10)[0]:
        wfs.extend([w_low_samples[transition]+2])

    wrs_a = np.array(wrs)
    wfs_a = np.array(wfs)
     ###                                       Starting samples for each recording
    wrs_start_a = [wrs[0]]
    wfs_start_a = [wfs[0]]

    for transition in np.where(np.diff(wrs)>50000)[0]:
        wrs_start_a.extend([wrs[transition]])
    for transition in np.where(np.diff(wfs)>50000)[0]:
        wfs_start_a.extend([wfs[transition]])    
    #                                     Ending Samples for each Recording
    wrs_end_a = []
    wfs_end_a = []

    for transition in np.where(np.diff(wrs)>30000)[0]:
        wrs_end_a.extend([wrs[transition]])
    for transition in np.where(np.diff(wfs)>30000)[0]:
        wfs_end_a.extend([wfs[transition]])

    wrs_end_a.append(wrs[-1])
    wfs_end_a.append(wfs[-1])
    #                              Visualization only Each Recordings Starting sample and ending sampling
    wrs_s = [] 
    wfs_s = []

    for num in num_rec_done:
        if (num <= rec_num):
            wrs_s.extend([wrs.index(wrs_start_a[num_rec_done[num]])]) 
    for num in num_rec_done:
        if (num <= rec_num):
            wfs_s.extend([wfs.index(wfs_start_a[num_rec_done[num]])]) 
    ##
    wrs_s_a = np.array(wrs_s)
    wfs_s_a = np.array(wfs_s)
    ##
    wrs_e = []
    wfs_e = []

    for num in num_rec_done:
        if (num <= rec_num):
            wrs_e.extend([wrs.index(wrs_end_a[num_rec_done[num]])]) 
    for num in num_rec_done:
        if (num <= rec_num):
            wfs_e.extend([wfs.index(wfs_end_a[num_rec_done[num]])]) 

    wrs_e_a = np.array(wrs_e)
    wfs_e_a = np.array(wfs_e)
    #creating a list, in which ers_x[0] contains the entire length of recording 1, and ers_x[1] contains all of recording 2
    wrs_i = []
    wfs_i = []

    for num in num_rec_done:
        if (num <= rec_num):
            wrs_i.extend([[(wrs_s_a[num_rec_done[num]]),(wrs_e_a[num_rec_done[num]])]]) 
    for num in num_rec_done:
        if (num <= rec_num):
            wfs_i.extend([[(wfs_s_a[num_rec_done[num]]),(wfs_e_a[num_rec_done[num]])]]) 
    wrs_i_a = np.array(wrs_i)
    wfs_i_a = np.array(wfs_i)       

    wrs_x = []
    wfs_x = []

    for num in num_rec_done:
        if (num <= rec_num):
            wrs_x.extend([wrs_a[wrs_i_a[num][0]+1:wrs_i_a[num][1]+1]])
    for num in num_rec_done:
        if (num <= rec_num):
            wfs_x.extend([wfs_a[wfs_i_a[num][0]+1:wfs_i_a[num][1]+1]])
    
    # print(wrs_x[5])
    for num in num_rec_done:
        if (num <= rec_num):
            print('Recording ' + str(num+1))
            print('')
            print('wrs_x: ' + str(wrs_x[num][0]) +'   '+ str(wrs_x[num][-1])+'   '+str(np.shape(wrs_x[num]))+'   '+str(wfs_x[num][0]-wrs_x[num][0]))
            print('wfs_x: ' + str(wfs_x[num][0]) +'   '+ str(wfs_x[num][-1])+'   '+str(np.shape(wfs_x[num]))+'   '+str(wfs_x[num][-1]-wrs_x[num][-1]))
            print('')

def trigger_timestamps(path,rec_done,tc):
    
    #Pulls Wanted Data from directory location, turns it into session
    directory = path
    session = Session(directory)
    recordnode = session.recordnodes[0]
    data = session.recordnodes[0].recordings[0].continuous
    sampling_rate = 30000.
    analog_channels = data[0].samples
    times = data[0].timestamps/sampling_rate
    metadata = data[0].metadata  
    ##
    num_rec_done = list(range(rec_done))
    num_rec_done_a = np.array(num_rec_done)
    rec_num = rec_done
    # ERS
    np.shape(np.where(analog_channels[:,tc]>15000)[0]) 
    
    e_high_samples = np.where(analog_channels[:,tc]<15000)[0]
    e_low_samples = np.where(analog_channels[:,tc]>15000)[0]
  
    ers = []
    efs = []
    for transition in np.where(np.diff(e_high_samples)>10)[0]:
        ers.extend([e_high_samples[transition]])
    for transition in np.where(np.diff(e_low_samples)>10)[0]:
        efs.extend([e_low_samples[transition]])

    ers_a = np.array(ers)
    efs_a = np.array(efs)
    
    # Creating Lists Containing Starting sample # for each recording inside ers
    ers_start_a = [ers_a[0]] #first rising edge of reach recording ers_a[0]
    efs_start_a = [efs_a[0]] #first falling edge of each recording efs_a[0]

    for transition in np.where(np.diff(ers_a)>50000)[0]:
        ers_start_a.extend([ers_a[transition]])
    for transition in np.where(np.diff(efs_a)>50000)[0]:
        efs_start_a.extend([efs_a[transition]])    

    # Creating Lists Containing Ending Sample # for each Recording inside ers
    ers_end_a = [] #last rising edge of each recording
    efs_end_a = [] #last falling edge of each recording

    for transition in np.where(np.diff(ers_a)>30000)[0]:
        ers_end_a.extend([ers_a[transition]])
    for transition in np.where(np.diff(efs_a)>30000)[0]:
        efs_end_a.extend([efs_a[transition]])

    ers_end_a.append(ers_a[-2])
    efs_end_a.append(efs_a[-1])
    # Creating Lists Containing Starting index locations in ers for each recording
    ers_s = [] 
    efs_s = []

    for num in num_rec_done:
        if (num <= rec_num):
            ers_s.extend([ers.index(ers_start_a[num_rec_done[num]])]) 
    for num in num_rec_done:
        if (num <= rec_num):
            efs_s.extend([efs.index(efs_start_a[num_rec_done[num]])]) 

    ers_s_a = np.array(ers_s)
    efs_s_a = np.array(efs_s)
    # 
    ers_e = []
    efs_e = []

    for num in num_rec_done:
        if (num <= rec_num):
            ers_e.extend([ers.index(ers_end_a[num_rec_done[num]])]) 
    for num in num_rec_done:
        if (num <= rec_num):
            efs_e.extend([efs.index(efs_end_a[num_rec_done[num]])]) 

    ers_e_a = np.array(ers_e)
    efs_e_a = np.array(efs_e)
    # creating a list, in which ers_x[0] contains the entire length of recording 1, and ers_x[1] contains all of recording 2
    ers_i = []
    efs_i = []

    for num in num_rec_done:
        if (num <= rec_num):
            ers_i.extend([[(ers_s_a[num_rec_done[num]]),(ers_e_a[num_rec_done[num]])]]) 
    for num in num_rec_done:
        if (num <= rec_num):
            efs_i.extend([[(efs_s_a[num_rec_done[num]]),(efs_e_a[num_rec_done[num]])]]) 
    ers_i_a = np.array(ers_i)
    efs_i_a = np.array(efs_i)       

    ers_x = []
    efs_x = []

    for num in num_rec_done:
        if (num <= rec_num):
            ers_x.extend([ers_a[ers_i_a[num][0]+1:ers_i_a[num][1]+1]])  # [ers_a[ers_i_a[num][0]+1:ers_i_a[num][1]]])
    for num in num_rec_done:
        if (num <= rec_num):
            efs_x.extend([efs_a[efs_i_a[num][0]+1:efs_i_a[num][1]+1]])  # [efs_a[efs_i_a[num][0]+1:efs_i_a[num][1]]])
    # Print Statements
    print('ers & efs')
    print('')
    print('ers: ' + str(ers[0]) +'   '+ str(ers[-1])+ '   '+ str(np.shape(ers)))
    print('efs: ' + str(efs[0]) +'   '+ str(efs[-1])+'    '+ str(np.shape(efs)))
    print('')


    for num in num_rec_done:
        if (num <= rec_num):
            print('Recording ' + str(num+1))
            print('')
            print('ers_x: ' + str(ers_x[num][0]) +'   '+ str(ers_x[num][-1])+'   '+str(np.shape(ers_x[num]))+'   '+str(efs_x[num][0]-ers_x[num][0]))
            print('efs_x: ' + str(efs_x[num][0]) +'   '+ str(efs_x[num][-1])+'   '+str(np.shape(ers_x[num]))+'   '+str(efs_x[num][0]-ers_x[num][0]))
            print('')